# -*- coding: utf-8 -*-
#
#  Project name: MXCuBE
#  https://github.com/mxcube
#
#  This file is part of MXCuBE software.
#
#  MXCuBE is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  MXCuBE is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU General Lesser Public License
#  along with MXCuBE. If not, see <http://www.gnu.org/licenses/>.

"""Exporter Client implementation"""

import logging

from .StandardClient import (
    ProtocolError,
    StandardClient,
)

__copyright__ = """ Copyright © 2019 by the MXCuBE collaboration """
__license__ = "LGPLv3+"

CMD_SYNC_CALL = "EXEC"
CMD_ASNC_CALL = "ASNC"
CMD_METHOD_LIST = "LIST"
CMD_PROPERTY_READ = "READ"
CMD_PROPERTY_WRITE = "WRTE"
CMD_PROPERTY_LIST = "PLST"
CMD_NAME = "NAME"
RET_ERR = "ERR:"
RET_OK = "RET:"
RET_NULL = "NULL"
EVENT = "EVT:"

PARAMETER_SEPARATOR = "\t"
ARRAY_SEPARATOR = ""  # 0x001F


class ExporterClient(StandardClient):
    """ExporterClient class"""

    def on_message_received(self, msg):
        """Act if the message is an event, pass to StandardClient otherwise.
        Args:
            msg(str): The message.
        """
        if msg[:4] == EVENT:
            try:
                tokens = msg[4:].split(PARAMETER_SEPARATOR)
                self.on_event(tokens[0], tokens[1], int(tokens[2]))
            except Exception:
                logging.getLogger("HWR").exception("")
        else:
            StandardClient.on_message_received(self, msg)

    def get_method_list(self):
        """Get the list of the methods
        Returns:
            (list): List of strings (the methods)
        """
        cmd = CMD_METHOD_LIST
        ret = self.send_receive(cmd)
        ret = self.__process_return(ret)
        if ret is None:
            return None
        ret = ret.split(PARAMETER_SEPARATOR)
        if len(ret) > 1:
            if ret[-1] == "":
                ret = ret[0:-1]
        return ret

    def get_property_list(self):
        """Get the list of the properties.
        Returns:
            (list): List of strings (the properties)
        """
        cmd = CMD_PROPERTY_LIST
        ret = self.send_receive(cmd)
        ret = self.__process_return(ret)
        if ret is None:
            return None
        ret = ret.split(PARAMETER_SEPARATOR)
        if len(ret) > 1:
            if ret[-1] == "":
                ret = ret[0:-1]
        return ret

    def get_server_object_name(self):
        """Get the server object name
        Returns:
            (str): The name.
        """
        cmd = CMD_NAME
        ret = self.send_receive(cmd)
        return self.__process_return(ret)

    def execute(self, method, pars=None, timeout=-1):
        """Execute a command synchronous.
        Args:
            method(str): Method name
            pars(str): parameters
            timeout(float): Timeout [s]
        """

        # runScript returns results in a different way than any other
        # command, fix on MD side ?
        if method == "runScript" and pars is not None:
            pars = pars[0].split(",")

        cmd = "{} {} ".format(CMD_SYNC_CALL, method)

        if pars is not None:
            if isinstance(pars, (list, tuple)):
                for par in pars:
                    if isinstance(par, (list, tuple)):
                        par = self.create_array_parameter(par)
                    cmd += str(par) + PARAMETER_SEPARATOR
            else:
                cmd += str(pars)

        ret = self.send_receive(cmd, timeout)
        return self.__process_return(ret)

    def __process_return(self, ret):
        """Analyse the return message.
        Args:
            ret(str): Returned message
        Returns:
            (str): The stripped message or None
        Raises:
            ProtocolError
        """
        if ret[:4] == RET_ERR:
            msg = f"{self.get_server_object_name()} : {str(ret[4:])}"
            logging.getLogger("HWR").error(msg)
            raise Exception(ret[4:])
        if ret == RET_NULL:
            return None
        if ret[:4] == RET_OK:
            return ret[4:]
        raise ProtocolError

    def execute_async(self, method, pars=None):
        """Execute command asynchronous.
        Args:
            method(str): Method name
            pars(str): parameters
        """
        cmd = "{} {} ".format(CMD_ASNC_CALL, method)
        if pars is not None:
            for par in pars:
                cmd += str(par) + PARAMETER_SEPARATOR
        return self.send(cmd)

    def write_property(self, prop, value, timeout=-1):
        """Write property synchronous.
        Args:
            prop(str): property name
            value: sample, list or tuple
        """
        if isinstance(value, (list, tuple)):
            value = self.create_array_parameter(value)
        cmd = "{} {} {}".format(CMD_PROPERTY_WRITE, prop, str(value))
        ret = self.send_receive(cmd, timeout)
        return self.__process_return(ret)

    def read_property(self, prop, timeout=-1):
        """Read a property
        Args:
            prop(str): property name
        Returns:
            (str): reply from the process.
        """
        cmd = "{} {}".format(CMD_PROPERTY_READ, prop)
        ret = self.send_receive(cmd, timeout)
        process_return = None
        try:
            process_return = self.__process_return(ret)
        except Exception:
            logging.getLogger("HWR").exception("")
        return process_return

    def read_property_as_string_array(self, prop):
        """Read a property and convert the return value to list of strings.
        Args:
            prop(str): property name
        Returns:
            (list): List of strings
        """
        ret = self.read_property(prop)
        return self.parse_array(ret)

    def parse_array(self, value):
        """Parse to list
        Args:
            value(str): input string
        Returns:
            (list): List of strings
        """
        value = str(value)
        if value.startswith(ARRAY_SEPARATOR) is False:
            return None
        if value == ARRAY_SEPARATOR:
            return []
        value = value.lstrip(ARRAY_SEPARATOR).rstrip(ARRAY_SEPARATOR)
        return value.split(ARRAY_SEPARATOR)

    def create_array_parameter(self, value):
        """Create a string to send.
        Args:
            value: simple, tuple or list
        Returns:
            (str): formatted string
        """
        ret = ARRAY_SEPARATOR
        if value is not None:
            if isinstance(value, (list, tuple)):
                for item in value:
                    ret += str(item) + ARRAY_SEPARATOR
            else:
                ret += str(value)
        return ret

    def on_event(self, name, value, timestamp):
        """Action"""



"""
File:  PX1Collect.py

"""

import logging
import os
import socket
import subprocess
import sys
import time

import gevent
from SOLEILMergeImage import merge as merge_images

from mxcubecore import HardwareRepository as HWR
from mxcubecore.BaseHardwareObjects import HardwareObject
from mxcubecore.Command.Tango import DeviceProxy
from mxcubecore.HardwareObjects.abstract.AbstractCollect import AbstractCollect
from mxcubecore.TaskUtils import task

__author__ = "Vicente Rey Bakaikoa"
__credits__ = ["MXCuBE collaboration"]
__version__ = "2.3"


class PX1Collect(AbstractCollect, HardwareObject):
    """Main data collection class. Inherited from AbstractMulticollect
    Collection is done by setting collection parameters and
    executing collect command
    """

    adxv_host = "127.0.0.1"
    adxv_port = 8100
    adxv_interval = 2.0  # minimum time (in seconds) between image refresh on adxv

    goimg_dir = "/nfs/ruche/share-temp/Proxima/.goimgpx1"
    goimg_filename = "goimg.db"

    characterization_nb_merged_images = 10

    def __init__(self, name):
        """

        :param name: name of the object
        :type name: string
        """

        AbstractCollect.__init__(self)
        HardwareObject.__init__(self, name)

        self._error_msg = ""
        self.owner = None
        self.osc_id = None
        self._collecting = None

        self.mxlocal = None

        self.helical_positions = None

    def init(self):
        """
        Init method
        """

        self.collect_devname = self.get_property("tangoname")
        self.collect_device = DeviceProxy(self.collect_devname)

        self.collect_state_chan = self.get_channel_object("state")

        self.px1env_hwobj = self.get_object_by_role("environment")

        self.frontend_hwobj = self.get_object_by_role("frontend")

        self.lightarm_hwobj = self.get_object_by_role("lightarm")

        self.mxlocal_object = self.get_object_by_role("beamline_configuration")

        self.img2jpeg = self.get_property("imgtojpeg")
        undulators = self.get_undulators()

        self.exp_type_dict = {"Mesh": "raster", "Helical": "Helical"}

        det_px, det_py = HWR.beamline.detector.get_pixel_size()
        beam_div_hor, beam_div_ver = HWR.beamline.beam.get_beam_divergence()

        self.set_beamline_configuration(
            synchrotron_name="SOLEIL",
            directory_prefix=self.get_property("directory_prefix"),
            default_exposure_time=HWR.beamline.detector.get_default_exposure_time(),
            minimum_exposure_time=HWR.beamline.detector.get_minimum_exposure_time(),
            detector_fileext=HWR.beamline.detector.get_file_suffix(),
            detector_type=HWR.beamline.detector.get_detector_type(),
            detector_manufacturer=HWR.beamline.detector.get_manufacturer(),
            detector_model=HWR.beamline.detector.get_model(),
            detector_px=det_px,
            detector_py=det_py,
            undulators=undulators,
            focusing_optic=self.get_property("focusing_optic"),
            monochromator_type=self.get_property("monochromator"),
            beam_divergence_vertical=beam_div_ver,
            beam_divergence_horizontal=beam_div_hor,
            polarisation=self.get_property("polarisation"),
            input_files_server=self.get_property("input_files_server"),
        )

        self.emit("collectConnected", (True,))
        self.emit("collectReady", (True,))

    def data_collection_hook(self):
        """Main collection hook"""

        collection_type = self.current_dc_parameters["experiment_type"]

        self.log.info(
            "PX1Collect: Running PX1 data collection hook. Type is %s" % collection_type
        )

        if self.aborted_by_user:
            self.emit_collection_failed("Aborted by user")
            self.aborted_by_user = False
            return

        ready = self.prepare_devices_for_collection()

        if not ready:
            self.collection_failed("Cannot prepare collection")
            self.stop_collect()
            return

        if collection_type != "Characterization":  # standard
            prepare_ok = self.prepare_standard_collection()
        else:
            prepare_ok = self.prepare_characterization()

        self._collecting = True

        osc_seq = self.current_dc_parameters["oscillation_sequence"][0]
        # for progressBar brick
        self.emit("progressInit", "Collection", osc_seq["number_of_images"])

        #
        # Run
        #

        self.prepare_directories()

        if collection_type != "Characterization":  # standard
            self.start_standard_collection()
            self.follow_collection_progress()
        else:
            self.start_characterization()

        # includes

        self.data_collection_end()
        self.collection_finished()

    def prepare_standard_collection(self):
        osc_seq = self.current_dc_parameters["oscillation_sequence"][0]
        fileinfo = self.current_dc_parameters["fileinfo"]
        basedir = fileinfo["directory"]

        self.log.info("PX1Collect: fileinfo is %s " % str(fileinfo))
        imgname = fileinfo["template"] % osc_seq["start_image_number"]

        # move omega to start angle
        start_angle = osc_seq["start"]

        nb_images = osc_seq["number_of_images"]
        osc_range = osc_seq["range"]
        exp_time = osc_seq["exposure_time"]

        self.log.info(
            "PX1Collect:  nb_images: %s / osc_range: %s / exp_time: %s"
            % (nb_images, osc_range, exp_time)
        )

        self.collect_device.exposurePeriod = exp_time
        self.collect_device.numberOfImages = nb_images
        self.collect_device.imageWidth = osc_range

        # self.collect_device.collectAxis = "Omega"

        self.collect_device.startAngle = start_angle
        self.collect_device.triggerMode = 2

        self.collect_device.imagePath = basedir
        self.collect_device.imageName = imgname
        time.sleep(0.2)
        self.collect_device.PrepareCollect()
        ret = self.wait_collect_standby()
        if ret is False:
            logging.getLogger("user_level_log").info(
                "Collect server prepare error. Aborted"
            )
            return False

        self.prepare_headers()

        return True

    def start_standard_collection(self):
        self.emit("collectStarted", (self.owner, 1))
        HWR.beamline.detector.start_collection()
        self.collect_device.Start()

    def follow_collection_progress(self):
        osc_seq = self.current_dc_parameters["oscillation_sequence"][0]
        fileinfo = self.current_dc_parameters["fileinfo"]
        basedir = fileinfo["directory"]
        archive_dir = fileinfo["archive_directory"]
        template = fileinfo["template"]

        jpeg_template = os.path.splitext(template)[0] + ".jpeg"
        thumb_template = os.path.splitext(template)[0] + ".thumb.jpeg"

        osc_range = osc_seq["range"]
        osc_start = osc_seq["start"]
        nb_images = osc_seq["number_of_images"]

        first_imgno = osc_seq["start_image_number"]
        first_image_fullpath = os.path.join(basedir, template % first_imgno)
        first_image_jpegpath = os.path.join(archive_dir, jpeg_template % first_imgno)
        first_image_thumbpath = os.path.join(archive_dir, thumb_template % first_imgno)

        last_imgno = first_imgno + nb_images - 1
        last_image_fullpath = os.path.join(basedir, template % last_imgno)
        last_image_jpegpath = os.path.join(archive_dir, jpeg_template % last_imgno)
        last_image_thumbpath = os.path.join(archive_dir, thumb_template % last_imgno)

        # wait for first image
        self.adxv_latest_refresh = 0
        self.is_firstimg = True

        self.file_waiting_display = first_image_fullpath  # for showing on adxv
        self.wait_image_on_disk(first_image_fullpath)
        thumbs_up = self.generate_thumbnails(
            first_image_fullpath, first_image_jpegpath, first_image_thumbpath
        )
        if thumbs_up:
            self._store_image_in_lims(first_imgno)

        # update display while collect is running
        while self.is_moving() or self.is_firstimg:
            self.adxv_show_latest(fileinfo)
            time.sleep(0.1)

        # wait for last image
        self.wait_image_on_disk(last_image_fullpath)
        thumbs_up = self.generate_thumbnails(
            last_image_fullpath, last_image_jpegpath, last_image_thumbpath
        )
        self.adxv_sync_image(first_image_fullpath)
        if thumbs_up:
            self._store_image_in_lims(last_imgno)

    def prepare_characterization(self):
        osc_seq = self.current_dc_parameters["oscillation_sequence"][0]
        fileinfo = self.current_dc_parameters["fileinfo"]

        basedir = fileinfo["directory"]

        merged_images = self.characterization_nb_merged_images

        osc_range = float(osc_seq["range"]) / merged_images
        exp_time = float(osc_seq["exposure_time"]) / merged_images

        self.collect_device.numberOfImages = int(merged_images)
        self.collect_device.exposurePeriod = exp_time
        self.collect_device.imagePath = basedir
        self.collect_device.imageWidth = osc_range

        HWR.beamline.detector.set_image_headers(["Angle_increment %.4f" % osc_range])

        return True

    def start_characterization(self):
        osc_seq = self.current_dc_parameters["oscillation_sequence"][0]
        fileinfo = self.current_dc_parameters["fileinfo"]

        start_angle = osc_seq["start"]
        nb_images = osc_seq["number_of_images"]
        merged_images = self.characterization_nb_merged_images
        wedge_range = osc_seq["range"] / merged_images

        image_template = fileinfo["template"]

        oscil_image_template = image_template[:3] + "_wdg" + image_template[3:]

        start = start_angle

        self.emit("collectStarted", (self.owner, 1))

        for imgno in range(nb_images):
            image_filename = image_template % (imgno + 1)
            image_fullpath = os.path.join(fileinfo["directory"], image_filename)

            series_first_image = oscil_image_template % (
                (start - start_angle) / wedge_range + 1
            )
            series_last_image = oscil_image_template % (
                (start - start_angle) / wedge_range + merged_images
            )
            series_last_fullpath = os.path.join(
                fileinfo["directory"], series_last_image
            )

            logging.info(
                "PX1Collect: start image series for angle %s - first image name is %s"
                % (start, series_first_image)
            )

            self.collect_device.imageName = series_first_image
            self.collect_device.startAngle = start
            self.wait_collect_ready()
            self.collect_device.prepareCollect()
            self.wait_collect_ready()

            HWR.beamline.detector.set_image_headers(["Start_angle %.4f" % start])
            self.collect_device.Start()

            # file names to wait for

            # wait for completion
            self.wait_collect_ready()
            logging.info(
                "PX1Collect:   waiting to finish image series for angle %s - waiting for %s"
                % (start, series_last_fullpath)
            )
            self.wait_image_on_disk(series_last_fullpath)
            self.adxv_sync_image(series_last_fullpath)

            # merge images into one image
            merge_images(
                series_first_image,
                image_filename,
                fileinfo["directory"],
                start,
                osc_seq["range"],
                osc_seq["exposure_time"],
            )

            logging.info(
                "PX1Collect:   waiting for merged image to appear on disk %s - waiting for %s"
                % (start, image_fullpath)
            )
            self.wait_image_on_disk(image_fullpath)
            self.adxv_sync_image(image_fullpath)

            start += 90.0

    def data_collection_end(self):
        #
        # data collection end (or abort)
        #
        self.log.info("PX1Collect: finishing data collection ")
        HWR.beamline.diffractometer.omega.stop()
        HWR.beamline.fast_shutter.closeShutter()

        self.emit("progressStop")

    def data_collection_failed(self):
        self.log.info(
            "PX1Collect: Data collection failed. recovering sequence should go here"
        )

    def collect_finished(self, green):
        logging.info("PX1Collect: Data collection finished")

    def collect_failed(self, par):
        logging.exception("PX1Collect: Data collection failed")
        self.current_dc_parameters["status"] = "failed"
        exc_type, exc_value, exc_tb = sys.exc_info()
        failed_msg = "Data collection failed!\n%s" % exc_value
        self.emit(
            "collectOscillationFailed",
            (
                self.owner,
                False,
                failed_msg,
                self.current_dc_parameters.get("collection_id"),
                1,
            ),
        )

        HWR.beamline.detector.stop_collection()
        HWR.beamline.diffractometer.omega.stop()
        self.data_collection_end()

    def set_helical_pos(self, arg):
        """
        Descript. : 8 floats describe
        p1AlignmY, p1AlignmZ, p1CentrX, p1CentrY
        p2AlignmY, p2AlignmZ, p2CentrX, p2CentrY
        """
        self.helical_positions = [
            arg["1"]["phiy"],
            arg["1"]["phiz"],
            arg["1"]["sampx"],
            arg["1"]["sampy"],
            arg["2"]["phiy"],
            arg["2"]["phiz"],
            arg["2"]["sampx"],
            arg["2"]["sampy"],
        ]

    def setMeshScanParameters(self, num_lines, num_images_per_line, mesh_range):
        """
        Descript. :
        """
        pass

    def trigger_auto_processing(self, *args):
        pass

    ## generate snapshots and data thumbnails ##
    @task
    def _take_crystal_snapshot(self, filename):
        """
        Descript. :
        """
        if not self.is_sampleview_phase():
            self.go_to_sampleview()
            time.sleep(2)  # allow time to refresh display after

        self.lightarm_hwobj.adjustLightLevel()
        time.sleep(0.3)  # allow time to refresh display after

        HWR.beamline.sample_view.save_snapshot(filename)
        self.log.debug("PX1Collect:  - snapshot saved to %s" % filename)

    def generate_thumbnails(self, filename, jpeg_filename, thumbnail_filename):
        #
        # write info on LIMS

        try:
            logging.info("PX1Collect: Generating thumbnails for %s" % filename)
            logging.info("PX1Collect:       jpeg file: %s" % jpeg_filename)
            logging.info("PX1Collect:  thumbnail file: %s" % thumbnail_filename)

            self.wait_image_on_disk(filename)
            if os.path.exists(filename):
                subprocess.Popen([self.img2jpeg, filename, jpeg_filename, "0.4"])
                subprocess.Popen([self.img2jpeg, filename, thumbnail_filename, "0.1"])
                return True
            else:
                logging.info(
                    "PX1Collect: Oopps.  Trying to generate thumbs but  image is not on disk"
                )
                return False
        except Exception:
            logging.error("PX1Collect: Cannot generate thumbnails for %s" % filename)
            logging.exception("")
            return False

    ## generate snapshots and data thumbnails (END) ##

    ## FILE SYSTEM ##
    def wait_image_on_disk(self, filename, timeout=20.0):
        start_wait = time.time()
        while not os.path.exists(filename):
            if time.time() - start_wait > timeout:
                logging.info("PX1Collect: Giving up waiting for image. Timeout")
                break
            time.sleep(0.1)
        logging.info(
            "PX1Collect: Waiting for image %s ended in  %3.2f secs"
            % (filename, time.time() - start_wait)
        )

    def check_directory(self, basedir):
        if not os.path.exists(basedir):
            try:
                os.makedirs(basedir)
            except OSError as e:
                if e.errno != errno.EEXIST:
                    raise

    def prepare_directories(self):
        fileinfo = self.current_dc_parameters["fileinfo"]
        basedir = fileinfo["directory"]
        process_dir = basedir.replace("RAW_DATA", "PROCESSED_DATA")

        try:
            os.chmod(process_dir, 0o777)
        except Exception:
            self.log.error(
                "PX1Collect: Error changing permissions for PROCESS directory"
            )
            self.log.exception()

        self.create_goimg_file(process_dir)

    def create_goimg_file(self, dirname):
        db_f = os.path.join(self.goimg_dir, self.goimg_filename)
        if os.path.exists(db_f):
            os.remove(db_f)

        with open(db_f, "w") as fd:
            fd.write(dirname)
        os.chmod(db_f, 0o777)

    def create_file_directories(self):
        """
        Method create directories for raw files and processing files.
        Directorie for xds.input and auto_processing are created
        """
        self.create_directories(
            self.current_dc_parameters["fileinfo"]["directory"],
            self.current_dc_parameters["fileinfo"]["process_directory"],
        )

        """create processing directories and img links"""
        xds_directory, auto_directory = self.prepare_input_files()
        try:
            # self.create_directories(xds_directory, auto_directory)
            # os.system("chmod -R 777 %s %s" % (xds_directory, auto_directory))
            """todo, create link of imgs for auto_processing
            try:
                os.symlink(files_directory, os.path.join(process_directory, "img"))
            except os.error, e:
                if e.errno != errno.EEXIST:
                    raise
            """
            pass
            # os.symlink(files_directory, os.path.join(process_directory, "img"))
        except Exception:
            logging.exception("PX1Collect: Could not create processing file directory")
            return

        if xds_directory:
            self.current_dc_parameters["xds_dir"] = xds_directory

        if auto_directory:
            self.current_dc_parameters["auto_dir"] = auto_directory

    def prepare_input_files(self):
        """
        Descript. :
        """
        i = 1
        log = logging.getLogger("user_level_log")
        while True:
            xds_input_file_dirname = "xds_%s_%s_%d" % (
                self.current_dc_parameters["fileinfo"]["prefix"],
                self.current_dc_parameters["fileinfo"]["run_number"],
                i,
            )
            xds_directory = os.path.join(
                self.current_dc_parameters["fileinfo"]["process_directory"],
                xds_input_file_dirname,
            )
            if not os.path.exists(xds_directory):
                break
            i += 1
        self.current_dc_parameters["xds_dir"] = xds_directory

        mosflm_input_file_dirname = "mosflm_%s_run%s_%d" % (
            self.current_dc_parameters["fileinfo"]["prefix"],
            self.current_dc_parameters["fileinfo"]["run_number"],
            i,
        )
        mosflm_directory = os.path.join(
            self.current_dc_parameters["fileinfo"]["process_directory"],
            mosflm_input_file_dirname,
        )

        log.info("  - xds: %s / mosflm: %s" % (xds_directory, mosflm_directory))
        return xds_directory, mosflm_directory

    ## FILE SYSTEM (END) ##

    def prepare_devices_for_collection(self):
        fileinfo = self.current_dc_parameters["fileinfo"]
        basedir = fileinfo["directory"]

        self.check_directory(basedir)

        # check fast shutter closed. others opened
        shutok = self.check_shutters()
        shutok = True

        if not shutok:
            logging.getLogger("user_level_log").info(
                " Shutters not ready for collection. Aborted"
            )
            return False

        detok = HWR.beamline.detector.prepare_collection(self.current_dc_parameters)
        if not detok:
            logging.getLogger("user_level_log").info(
                "Cannot prepare detector for collection. Aborted"
            )
            return False

        adxv_ok = self.adxv_connect()

        diff_ok = self.diffractometer_prepare_collection()
        if not diff_ok:
            logging.getLogger("user_level_log").info(
                "Cannot prepare diffractometer for collection. Aborted"
            )
            return False

        return True

    def diffractometer_prepare_collection(self):
        HWR.beamline.diffractometer.wait_device_ready(timeout=10)

        # go to collect phase
        if not self.is_collect_phase():
            success = self.go_to_collect()
            if not success:
                self.log.info("PX1Collect: Cannot set COLLECT phase")
                return False
        return True

    def prepare_headers(self):
        osc_seq = self.current_dc_parameters["oscillation_sequence"][0]

        ax, ay, bx, by = self.get_beam_configuration()

        dist = HWR.beamline.detector.distance.get_value()
        wavlen = HWR.beamline.energy.get_wavelength()

        start_angle = osc_seq["start"]
        nb_images = osc_seq["number_of_images"]
        img_range = osc_seq["range"]
        exp_time = osc_seq["exposure_time"]

        kappa_angle = HWR.beamline.diffractometer.kappa.get_value()

        _settings = [
            ["Wavelength %.5f", wavlen],
            ["Detector_distance %.4f", dist / 1000.0],
            ["Beam_x %.2f", ax * dist + bx],
            ["Beam_y %.2f", ay * dist + by],
            ["Alpha %.2f", 49.64],
            ["Start_angle %.4f", start_angle],
            ["Angle_increment %.4f", img_range],
            ["Oscillation_axis %s", "Omega"],
            ["Detector_2theta %.4f", 0.0],
            ["Polarization %.3f", 0.990],
            ["Kappa %.4f", kappa_angle],
        ]

        # if self.oscaxis == "Phi":
        # _settings.append(["Chi %.4f", self.omega_hwo.get_value()])
        # _settings.append(["Phi %.4f", start])
        # elif self.oscaxis == "Omega":
        _settings.append(
            ["Phi %.4f", HWR.beamline.diffractometer.kappa_phi.get_value()]
        )
        _settings.append(["Chi %.4f", start_angle])
        HWR.beamline.detector.set_image_headers(_settings)

    def check_shutters(self):
        # Check safety shutter
        if self.check_shutter_opened(
            HWR.beamline.safety_shutter, "Safety shutter"
        ) and self.check_shutter_opened(self.frontend_hwobj, "Front end"):
            return True
        else:
            return False

    def check_shutter_opened(self, shut_hwo, shutter_name="shutter"):
        if shut_hwo.isShutterOpened():
            return True

        if shut_hwo.get_state() == "disabled":
            logging.getLogger("user_level_log").warning(
                "%s disabled. Collect cancelled" % shutter_name
            )
            return False
        elif shut_hwo.get_state() in ["fault", "alarm", "error"]:
            logging.getLogger("user_level_log").warning(
                "%s is in fault state. Collect cancelled" % shutter_name
            )
            return False
        elif shut_hwo.isShutterClosed():
            shut_hwo.openShutter()
            return shut_hwo.waitShutter("opened")
        else:
            logging.getLogger("user_level_log").warning(
                "%s is in an unhandled state. Collect cancelled" % shutter_name
            )
            return False

    def close_fast_shutter(self):
        HWR.beamline.fast_shutter.closeShutter()

    def close_safety_shutter(self):
        pass

    ## COLLECT SERVER STATE ##
    def wait_collect_standby(self, timeout=10):
        t0 = time.time()
        while not self.is_standby():
            elapsed = time.time() - t0
            if elapsed > timeout:
                break
            time.sleep(0.1)

    def wait_collect_moving(self, timeout=10):
        t0 = time.time()
        while not self.is_moving():
            elapsed = time.time() - t0
            if elapsed > timeout:
                break
            time.sleep(0.1)

    def wait_collect_ready(self, timeout=10):
        t0 = time.time()
        while self.is_moving():
            elapsed = time.time() - t0
            if elapsed > timeout:
                break
            time.sleep(0.1)

    def is_standby(self):
        return str(self.collect_state_chan.get_value()) == "STANDBY"

    def is_moving(self):
        return str(self.collect_state_chan.get_value()) in ["MOVING", "RUNNING"]

    ## COLLECT SERVER STATE (END) ##

    ## PX1 ENVIRONMENT PHASE HANDLING ##
    def is_collect_phase(self):
        return self.px1env_hwobj.isPhaseCollect()

    def go_to_collect(self, timeout=180):
        self.px1env_hwobj.gotoCollectPhase()
        gevent.sleep(0.5)

        t0 = time.time()
        while True:
            env_state = self.px1env_hwobj.get_state()
            if env_state != "RUNNING" and self.is_collect_phase():
                break
            if time.time() - t0 > timeout:
                self.log.debug(
                    "PX1Collect: timeout sending supervisor to collect phase"
                )
                break
            gevent.sleep(0.5)

        return self.px1env_hwobj.isPhaseCollect()

    def is_sampleview_phase(self):
        return self.px1env_hwobj.isPhaseVisuSample()

    def go_to_sampleview(self, timeout=180):
        self.px1env_hwobj.gotoSampleViewPhase()

        gevent.sleep(0.5)

        t0 = time.time()
        while True:
            env_state = self.px1env_hwobj.get_state()
            if env_state != "RUNNING" and self.is_sampleview_phase():
                break
            if time.time() - t0 > timeout:
                self.log.debug(
                    "PX1Collect: timeout sending supervisor to sample view phase"
                )
                break
            gevent.sleep(0.5)

        self.lightarm_hwobj.adjustLightLevel()
        return self.is_sampleview_phase()

    ## PX1 ENVIRONMENT PHASE HANDLING (END) ##

    @task
    def move_motors(self, motor_position_dict):
        """
        Descript. :
        """
        HWR.beamline.diffractometer.move_motors(motor_position_dict)

    def get_undulators_gaps(self):
        """
        Descript. : return gaps as dict. In our case we have one gap,
                    others are 0
        """
        if HWR.beamline.energy:
            try:
                u20_gap = HWR.beamline.energy.getCurrentUndulatorGap()
                return {"u20": u20_gap}
            except Exception:
                return {}
        else:
            return {}

    def get_slit_gaps(self):
        """
        Descript. :
        """
        if HWR.beamline.beam is not None:
            return HWR.beamline.beam.get_slits_gap()
        return None, None

    def get_beam_shape(self):
        """
        Descript. :
        """
        if HWR.beamline.beam is not None:
            return HWR.beamline.beam.get_beam_shape()

    def get_machine_current(self):
        """
        Descript. :
        """
        if HWR.beamline.machine_info:
            return HWR.beamline.machine_info.get_current()
        else:
            return 0

    def get_machine_message(self):
        """
        Descript. :
        """
        if HWR.beamline.machine_info:
            return HWR.beamline.machine_info.get_message()
        else:
            return ""

    def get_machine_fill_mode(self):
        """
        Descript. :
        """
        if HWR.beamline.machine_info:
            return HWR.beamline.machine_info.get_fill_mode()
        else:
            return ""

    def get_beam_configuration(self):
        pars_beam = self.mxlocal_object["SPEC_PARS"]["beam"]
        ax = pars_beam.get_property("ax")
        ay = pars_beam.get_property("ay")
        bx = pars_beam.get_property("bx")
        by = pars_beam.get_property("by")
        return [ax, ay, bx, by]

    def get_undulators(self):
        return [U20()]

    ## OTHER HARDWARE OBJECTS (END) ##

    ## ADXV display images ##
    def adxv_connect(self):
        #  connect every time?? maybe we can do better
        try:
            res = socket.getaddrinfo(
                self.adxv_host, self.adxv_port, 0, socket.SOCK_STREAM
            )
            af, socktype, proto, canonname, sa = res[0]
            self.adxv_socket = socket.socket(af, socktype, proto)
            self.adxv_socket.connect((self.adxv_host, self.adxv_port))
            logging.getLogger().info("PX1Collect: ADXV Visualization connected.")
        except Exception:
            self.adxv_socket = None
            logging.getLogger().info("PX1Collect: WARNING: Can't connect to ADXV.")

    def adxv_show_latest(self, fileinfo):
        now = time.time()
        elapsed = now - self.adxv_latest_refresh

        template = fileinfo["template"]
        basedir = fileinfo["directory"]

        if self.file_waiting_display is not None:
            if os.path.exists(self.file_waiting_display):
                self.adxv_sync_image(self.file_waiting_display)
                self.file_waiting_display = None
                self.is_firstimg = False

        # find next file to display
        if elapsed >= self.adxv_interval and self.file_waiting_display is None:
            _current_img_no = self.collect_device.currentImageSpi
            self.file_waiting_display = os.path.join(
                basedir, template % _current_img_no
            )
            self.adxv_last_refresh = time.time()

    def adxv_sync_image(self, filename):
        adxv_send_cmd = "\nload_image %s\n" + chr(32)

        try:
            if not self.adxv_socket:
                try:
                    self.adxv_connect()
                except Exception as err:
                    self.adxv_socket = None
                    logging.info(
                        "PX1Collect: ADXV: Warning: Can't connect to adxv socket to follow collect."
                    )
                    logging.error("PX1Collect: ADXV0: msg= %s" % err)
            else:
                logging.info(("PX1Collect: ADXV: " + adxv_send_cmd[1:-2]) % imgname)
                self.adxv_socket.send(adxv_send_cmd % imgname)
        except Exception:
            try:
                del self.adxv_socket
                self.adxv_connect()
            except Exception as err:
                self.adxv_socket = None
                logging.error("PX1Collect: ADXV1: msg= %s" % err)

    ## ADXV display images (END) ##


class U20(object):
    def __init__(self):
        self.type = "u20"


def test_hwo(hwo):
    print("PX1Environemnt (collect phase): ", hwo.is_collect_phase())
    print("Shutters (ready for collect): ", hwo.check_shutters())
    print("is collect? ", hwo.is_collect_phase())
    print("is samplevisu? ", hwo.is_sampleview_phase())
    print("goint to sample visu")
    hwo.go_to_sampleview()
    # print "goint to collect"
    # hwo.go_to_collect()
    print("is collect? ", hwo.is_collect_phase())
    print("is samplevisu? ", hwo.is_sampleview_phase())



#
#  Project name: MXCuBE
#  https://github.com/mxcube
#
#  This file is part of MXCuBE software.
#
#  MXCuBE is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  MXCuBE is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with MXCuBE. If not, see <http://www.gnu.org/licenses/>.

"""
EMBLBeamInfo
Hardware object is used to define final beam size and shape.
It can include aperture, slits and/or beam focusing hwobj
"""

import ast
import logging
import sys

from mxcubecore.HardwareObjects.abstract.AbstractBeam import AbstractBeam

__credits__ = ["EMBL Hamburg"]
__license__ = "LGPLv3+"
__category__ = "General"


class EMBLBeam(AbstractBeam):
    """Hardware object is used to define final beam size and shape"""

    def __init__(self, *args):
        """Defines all variables"""

        AbstractBeam.__init__(self, *args)

        self.focus_mode = None
        self.aperture_pos_name = None
        self.default_beam_divergence = None

        self.chan_beam_position_hor = None
        self.chan_beam_position_ver = None
        self.chan_beam_size_microns = None
        self.chan_beam_shape_ellipse = None

    def init(self):
        """Initialized all variables"""

        self._beam_info_dict["size_x"] = 9999
        self._beam_info_dict["size_y"] = 9999

        self._aperture = self.get_object_by_role("aperture")
        if self._aperture is not None:
            self.connect(
                self._aperture,
                "diameterIndexChanged",
                self.aperture_diameter_changed,
            )
        else:
            self.log.debug("BeamInfo: Aperture hwobj not defined")

        self._slits = self.get_object_by_role("slits")
        if self._slits is not None:
            self.connect(self._slits, "valueChanged", self.slits_gap_changed)
        else:
            self.log.debug("BeamInfo: Slits hwobj not defined")

        self._definer = self.get_object_by_role("beam_focusing")
        if self._definer is not None:
            (
                focus_mode_name,
                self._beam_size_dict["definer"],
            ) = self._definer.get_active_focus_mode()
            self.connect(
                self._definer,
                "focusingModeChanged",
                self.focusing_mode_changed,
            )
        else:
            self.log.debug("BeamInfo: Beam focusing hwobj not defined")

        self.chan_beam_position_hor = self.get_channel_object("BeamPositionHorizontal")
        if self.chan_beam_position_hor:
            self.chan_beam_position_hor.connect_signal(
                "update", self.beam_pos_hor_changed
            )
            self.beam_pos_hor_changed(self.chan_beam_position_hor.get_value())
        self.chan_beam_position_ver = self.get_channel_object("BeamPositionVertical")
        if self.chan_beam_position_ver:
            self.chan_beam_position_ver.connect_signal(
                "update", self.beam_pos_ver_changed
            )
        self.chan_beam_size_microns = self.get_channel_object("BeamSizeMicrons")
        self.chan_beam_shape_ellipse = self.get_channel_object("BeamShapeEllipse")
        self.default_beam_divergence = ast.literal_eval(
            self.get_property("defaultBeamDivergence")
        )

    def get_beam_divergence_hor(self):
        """Returns beam horizontal beam divergence

        :return: float
        """
        beam_divergence_hor = None

        if self._definer is not None:
            beam_divergence_hor = self._definer.get_divergence_hor()
        else:
            beam_divergence_hor = self.default_beam_divergence[0]

        return beam_divergence_hor

    def get_beam_divergence_ver(self):
        """Returns vertical beam divergence

        :return: float
        """
        beam_divergence_ver = None

        if self._definer is not None:
            beam_divergence_ver = self._definer.get_divergence_ver()
        else:
            beam_divergence_ver = self.default_beam_divergence[1]

        return beam_divergence_ver

    def beam_pos_hor_changed(self, value):
        """Updates horizontal beam position

        :param value: horizontal beam position
        :type value: float
        :return: None
        """
        self._beam_position_on_screen[0] = value
        self.emit("beamPosChanged", (self._beam_position_on_screen,))

    def beam_pos_ver_changed(self, value):
        """Updates vertical beam position

        :param value: vertical beam position
        :type value: float
        :return: None
        """
        self._beam_position_on_screen[1] = value
        self.emit("beamPosChanged", (self._beam_position_on_screen,))

    def get_beam_position(self):
        """Returns beam mark position

        :return: [float, float]
        """
        if self.chan_beam_position_hor and self.chan_beam_position_ver:
            self._beam_position_on_screen = [
                self.chan_beam_position_hor.get_value(),
                self.chan_beam_position_ver.get_value(),
            ]
        return self._beam_position_on_screen

    def set_beam_position(self, beam_x, beam_y):
        """Sets the beam mark position

        :param beam_x: horizontal beam mark position
        :type beam_x: int
        :param beam_y: vertical beam mark position
        :type beam_y: int
        :return: None
        """
        self._beam_position_on_screen = [beam_x, beam_y]

        if self.chan_beam_position_hor and self.chan_beam_position_ver:
            self.chan_beam_position_hor.set_value(int(beam_x))
            self.chan_beam_position_ver.set_value(int(beam_y))
        else:
            self.emit("beamPosChanged", (self._beam_position_on_screen,))

    def aperture_diameter_changed(self, name, size):
        """Method called when aperture changed. Evaluates beam mark position

        :param name: position name
        :type name: str
        :param size: diameter size
        :type size: float
        :return: None
        """
        self._beam_size_dict["aperture"] = [size, size]
        self.aperture_pos_name = name
        self.update_beam_info()
        self.re_emit_values()

    def get_aperture_pos_name(self):
        """Returns current aperture position

        :return: position name as str
        """
        return self._aperture.get_current_pos_name()

    def slits_gap_changed(self, size):
        """Method when slits gap changes. Evaluates beam mark position

        :param size: gap size
        :type size: list of two floats
        :return: None
        """
        self._beam_size_dict["slits"] = size
        self.update_beam_info()
        self.re_emit_values()

    def focusing_mode_changed(self, name, size):
        """Updates beam mark size when beam focusing changes

        :param name: focusing name
        :type name: str
        :param size: beam size
        :type size: float
        :return: None
        """
        self.focus_mode = name
        self._beam_size_dict["definer"] = size
        self.update_beam_info()
        self.re_emit_values()

    def get_beam_size(self):
        """Returns beam size in microns

        :return: Tuple(int, int)
        """
        self.update_beam_info()
        return (self._beam_info_dict["size_x"], self._beam_info_dict["size_y"])

    def get_beam_shape(self):
        """Returns beam shape

        :return: beam shape as str
        """
        self.update_beam_info()
        return self._beam_info_dict["shape"]

    def get_slits_gap(self):
        """Returns slits gap

        :return: list of two floats
        """
        self.update_beam_info()
        slits_gap = [None, None]

        if self._beam_size_dict["slits"] != [sys.float_info.max, sys.float_info.max]:
            slits_gap = self._beam_size_dict["slits"]

        return slits_gap

    def set_slits_gap(self, width_microns, height_microns):
        """Sets slits gaps

        :param width_microns: width in microns
        :type width_microns: float
        :param height_microns: height in microns
        :type height_microns: float
        :return: None
        """
        if self.focus_mode == "Double":
            logging.getLogger("GUI").warning(
                "Slits are disabled in the Double focus mode"
            )
        else:
            self._slits.set_horizontal_gap(width_microns / 1000.0)
            self._slits.set_vertical_gap(height_microns / 1000.0)

    def update_beam_info(self):
        """Called if aperture, slits or focusing has been changed

        :return: dictionary,{size_x:0.1, size_y:0.1, shape:"rectangular"}
        """
        size_x = min(
            self._beam_size_dict["aperture"][0],
            self._beam_size_dict["slits"][0],
            self._beam_size_dict["definer"][0],
        )
        size_y = min(
            self._beam_size_dict["aperture"][1],
            self._beam_size_dict["slits"][1],
            self._beam_size_dict["definer"][1],
        )

        if size_x == 9999 or size_y == 9999:
            return

        if (
            abs(size_x - self._beam_info_dict.get("size_x", 0)) > 1e-3
            or abs(size_y - self._beam_info_dict.get("size_y", 0)) > 1e-3
        ):
            self._beam_info_dict["size_x"] = size_x
            self._beam_info_dict["size_y"] = size_y

            if self._beam_size_dict["aperture"] <= [size_x, size_y]:
                self._beam_info_dict["shape"] = "ellipse"
            else:
                self._beam_info_dict["shape"] = "rectangular"

            if (
                self.chan_beam_size_microns is not None
                and self._beam_info_dict["size_x"] < 1.3
                and self._beam_info_dict["size_y"] < 1.3
            ):
                self.chan_beam_size_microns.set_value(
                    (
                        self._beam_info_dict["size_x"] * 1000,
                        self._beam_info_dict["size_y"] * 1000,
                    )
                )
            if self.chan_beam_shape_ellipse:
                self.chan_beam_shape_ellipse.set_value(
                    self._beam_info_dict["shape"] == "ellipse"
                )

    def re_emit_values(self):
        """Emits signals

        :return: None
        """
        if (
            self._beam_info_dict["size_x"] != sys.float_info.max
            and self._beam_info_dict["size_y"] != sys.float_info.max
        ):
            self.emit(
                "beamSizeChanged",
                (
                    (
                        self._beam_info_dict["size_x"] * 1000,
                        self._beam_info_dict["size_y"] * 1000,
                    ),
                ),
            )
            self.emit("beamInfoChanged", (self._beam_info_dict,))

    def get_beam_info(self):
        """Returns beam info

        :return: dict
        """
        self.update_beam_info()
        return self._beam_info_dict

    def re_emit_values(self):
        """Reemits all signals

        :return: None
        """
        self.emit("beamInfoChanged", (self._beam_info_dict,))
        self.emit("beamPosChanged", (self._beam_position_on_screen,))

    def move_beam(self, direction, step=1):
        """Moves beam mark

        :param direction: direction
        :type direction: str
        :param step: step in pixels
        :type step: int
        :return: None
        """
        if direction == "left":
            self.chan_beam_position_hor.set_value(
                self._beam_position_on_screen[0] - step
            )
        elif direction == "right":
            self.chan_beam_position_hor.set_value(
                self._beam_position_on_screen[0] + step
            )
        elif direction == "up":
            self.chan_beam_position_ver.set_value(
                self._beam_position_on_screen[1] - step
            )
        elif direction == "down":
            self.chan_beam_position_ver.set_value(
                self._beam_position_on_screen[1] + step
            )

    def get_focus_mode(self):
        """Returns current focusing mode

        :return: str
        """
        focus_mode = None

        if self._definer is not None:
            focus_mode = self._definer.get_focus_mode()

        return focus_mode



"""
Secure XML RPC request handler for workflow execution based on tokens.
Inspired from this code:
http://code.activestate.com/recipes/496786-simple-xml-rpc-server-over-https
"""

import sys

if sys.version_info > (3, 0):
    from xmlrpc.server import SimpleXMLRPCRequestHandler
else:
    from SimpleXMLRPCServer import SimpleXMLRPCRequestHandler


__author__ = "Olof Svensson"
__copyright__ = "Copyright 2012, ESRF"
__credits__ = ["MxCuBE collaboration"]

__version__ = ""
__maintainer__ = "Marcus Oskarsson"
__email__ = "marcus.oscarsson@esrf.fr"
__status__ = "Draft"


class SecureXMLRpcRequestHandler(SimpleXMLRPCRequestHandler):
    """
    Secure XML-RPC request handler class.

    It it very similar to SimpleXMLRPCRequestHandler but it checks for a
    "Token" entry in the header. If this token doesn't correspond to a
    reference token the server sends a "401" (Unauthorized) reply.
    """

    __referenceToken = None

    @staticmethod
    def setReferenceToken(token):
        SecureXMLRpcRequestHandler.__referenceToken = token

    def setup(self):
        self.connection = self.request
        self.rfile = self.connection.makefile("rb", self.rbufsize)
        self.wfile = self.connection.makefile("wb", self.wbufsize)

    def do_POST(self):
        """
        Handles the HTTPS POST request.

        It was copied out from SimpleXMLRPCServer.py and modified to check for "Token" in the headers.
        """
        # Check that the path is legal
        if not self.is_rpc_path_valid():
            self.report_404()
            return

        referenceToken = SecureXMLRpcRequestHandler.__referenceToken
        if (
            referenceToken is not None
            and "Token" in self.headers
            and referenceToken == self.headers["Token"]
        ):
            try:
                # Get arguments by reading body of request.
                # We read this in chunks to avoid straining
                # socket.read(); around the 10 or 15Mb mark, some platforms
                # begin to have problems (bug #792570).
                max_chunk_size = 10 * 1024 * 1024
                size_remaining = int(self.headers["content-length"])
                L = []
                while size_remaining:
                    chunk_size = min(size_remaining, max_chunk_size)
                    chunk = self.rfile.read(chunk_size)
                    if not chunk:
                        break
                    L.append(chunk.decode("utf-8"))
                    size_remaining -= len(L[-1])
                data = "".join(L)
                # In previous versions of SimpleXMLRPCServer, _dispatch
                # could be overridden in this class, instead of in
                # SimpleXMLRPCDispatcher. To maintain backwards compatibility,
                # check to see if a subclass implements _dispatch and dispatch
                # using that method if present.
                response = self.server._marshaled_dispatch(
                    data, getattr(self, "_dispatch", None)
                )
            except Exception:  # This should only happen if the module is buggy
                # internal error, report as HTTP server error
                self.send_response(500)
                logging.getLogger("HWR").exception("Exception during request.")

                self.end_headers()
            else:
                # got a valid XML RPC response
                self.send_response(200)
                self.send_header("Content-type", "text/xml")
                self.send_header("Content-length", str(len(response)))
                self.end_headers()
                self.wfile.write(response)

                # shut down the connection
                self.wfile.flush()
                self.connection.shutdown(1)
        else:
            # Unrecognized token - access unauthorized
            self.send_response(401)
            self.end_headers()



# encoding: utf-8
#
#  Project name: MXCuBE
#  https://github.com/mxcube
#
#  This file is part of MXCuBE software.
#
#  MXCuBE is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  MXCuBE is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with MXCuBE. If not, see <http://www.gnu.org/licenses/>.

"""
Mixin superclass for all mock actuators

Should be put as the first superclass,
e.g. class MotorMockup(ActuatorMockup, AbstractMotor):
"""

import random
import time

import gevent

from mxcubecore.HardwareObjects.abstract import AbstractActuator

__copyright__ = """ Copyright © 2010-2020 by the MXCuBE collaboration """
__license__ = "LGPLv3+"


class ActuatorMockup(AbstractActuator.AbstractActuator):
    """Mock Motor implementation"""

    def __init__(self, name):
        super().__init__(name)
        self.__move_task = None

    def init(self):
        """Initialisation method"""
        super().init()
        self.update_state(self.STATES.READY)

    def _move(self, value):
        """Simulated value change - override as needed

        Must set specific_state as needed, take a non-zero amount of time
        call update_value for intermediate positions
        and return the final value (in case it does not match the input value)

        Args:
            value : target actuator value

        Returns:
            final actuator value (may differ from target value)
        """
        time.sleep(random.uniform(0.1, 1.0))
        return value

    def get_value(self):
        """Read the actuator position.
        Returns:
            float: Actuator position.
        """
        return self._nominal_value

    def set_value(self, value, timeout=0):
        """
        Set actuator to absolute value.
        This is NOT the recommended way, but for technical reasons
        overriding is necessary in this particular case
        Args:
            value (float): target value
            timeout (float): optional - timeout [s],
                             If timeout == 0: return at once and do not wait (default);
                             if timeout is None: wait forever.
        Raises:
            ValueError: Value not valid or attempt to set read-only actuator.
            RuntimeError: Timeout.
        """
        if self.read_only:
            raise ValueError("Attempt to set value for read-only Actuator")
        if self.validate_value(value):
            self.update_state(self.STATES.BUSY)
            if timeout or timeout is None:
                with gevent.Timeout(
                    timeout, RuntimeError(f"Motor {self.username} timed out")
                ):
                    new_value = self._move(value)
                    self._set_value(new_value)
            else:
                self.__move_task = gevent.spawn(self._move, value)
                self.__move_task.link(self._callback)
        else:
            raise ValueError(f"Invalid value {value}; limits are {self.get_limits()}")

    def abort(self):
        """Immediately halt movement. By default self.stop = self.abort"""
        if self.__move_task is not None:
            self.__move_task.kill()
        self.update_state(self.STATES.READY)

    def _callback(self, move_task):
        value = move_task.get()
        if not isinstance(value, gevent.GreenletExit):
            self._set_value(value)

    def _set_value(self, value):
        """
        Implementation of specific set actuator logic.
        Args:
            value (float): target value
        """
        self.update_value(value)
        self.update_state(self.STATES.READY)



"""Custom MAXIV Beamline object.

Adds support for ``emulate`` configurable, which allows to specify
features to be emulated at the beamline.

Emulation of following features are supported:

* safety_shutter - emulate safety shutter opening
* detector_cover - emulate detector cover opening
* detector_motion - emulate detector distance moves

Example of ``emulate`` configuration::

  configuration:
    emulate:
      safety_shutter: true
      detector_cover: true
      detector_motion: true
"""

import mxcubecore.HardwareObjects.Beamline


class Beamline(mxcubecore.HardwareObjects.Beamline.Beamline):
    def emulate(self, feature: str) -> bool:
        """Check if some feature should be emulated."""
        emulate = self.get_property("emulate", {})
        return emulate.get(feature, False)



from datetime import datetime
from typing import Optional

from pydantic import (
    BaseModel,
    Field,
)


class CommonCollectionParamters(BaseModel):
    skip_existing_images: bool
    take_snapshots: int
    type: str
    label: str


class PathParameters(BaseModel):
    prefix: str
    subdir: str
    experiment_name: Optional[str]

    class Config:
        extra = "ignore"


class LegacyParameters(BaseModel):
    take_dark_current: int
    inverse_beam: bool
    num_passes: int
    overlap: float

    class Config:
        extra = "ignore"


class StandardCollectionParameters(BaseModel):
    num_images: int
    osc_start: Optional[float]
    osc_range: Optional[float]
    energy: float
    transmission: float
    resolution: float
    first_image: int
    kappa: Optional[float]
    kappa_phi: Optional[float]
    beam_size: float
    shutterless: bool
    selection: list = Field([])
    shape: str = ""

    class Config:
        extra = "ignore"


class BeamlineParameters(BaseModel):
    energy: float
    transmission: float
    resolution: float
    wavelength: float
    detector_distance: float
    beam_x: float
    beam_y: float
    beam_size_x: float
    beam_size_y: float
    beam_shape: str
    energy_bandwidth: float


class ISPYBCollectionParameters(BaseModel):
    flux_start: float
    flux_end: float
    start_time: datetime
    end_time: datetime
    chip_model: str
    mono_stripe: str
    number_of_rows: int
    number_of_columns: int



# encoding: utf-8
#
#  Project name: MXCuBE
#  https://github.com/mxcube.
#
#  This file is part of MXCuBE software.
#
#  MXCuBE is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  MXCuBE is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU General Lesser Public License
#  along with MXCuBE.  If not, see <http://www.gnu.org/licenses/>.

"""
Example xml_ configuration:

.. code-block:: xml

 <object class="ESRF.ESRFXRFSpectrum">
   <object href="/bliss" role="controller"/>
   <cfgfile>/users/blissadm/local/beamline_configuration/misc/15keV.cfg</cfgfile>
   <default_integration_time>3</default_integration_time>
   <default_energy_range>[2.0, 15]</default_energy_range>
   <cfg_energies>[15]</cfg_energies>
 </object>
"""

__copyright__ = """ Copyright © by the MXCuBE collaboration """
__license__ = "LGPLv3+"


import logging
import subprocess
import time
from datetime import datetime as dt
from pathlib import Path
from shutil import copy2
from zoneinfo import ZoneInfo

import numpy as np
from gevent import event, spawn

from mxcubecore import HardwareRepository as HWR
from mxcubecore.HardwareObjects.abstract.AbstractEnergyScan import AbstractEnergyScan


class GetStaticParameters:
    def __init__(self, config_file, element, edge):
        self.element = element
        self.edge = edge
        self.pars_dict = {}
        self.pars_dict = self._read_from_file(config_file)

    def _read_from_file(self, config_file):
        with Path(config_file).open("r") as fp:
            array = []
            for line in fp:
                if not line.startswith("#") and self.element in line:
                    array = line.split()
                    break

            try:
                static_pars = {}
                static_pars["atomic_nb"] = int(array[0])
                static_pars["eroi_min"] = float(array[11]) / 1000.0
                static_pars["eroi_max"] = float(array[12]) / 1000.0

                if "K" in self.edge:
                    th_energy = float(array[3]) / 1000.0
                elif "1" in self.edge:
                    # L1
                    th_energy = float(array[6]) / 1000.0
                elif "2" in self.edge:
                    # L2
                    th_energy = float(array[7]) / 1000.0
                else:
                    # L or L3
                    th_energy = float(array[8]) / 1000.0

                # all the values are in keV
                static_pars["edgeEnergy"] = th_energy
                static_pars["startEnergy"] = th_energy - 0.05
                static_pars["endEnergy"] = th_energy + 0.05
                static_pars["findattEnergy"] = th_energy + 0.03
                static_pars["remoteEnergy"] = th_energy + 1
            except TypeError:
                return {}
            return static_pars
        return {}


class ESRFEnergyScan(AbstractEnergyScan):
    def __init__(self, name):
        super().__init__(name)
        self.ctrl = None

    def execute_command(self, command_name, *args, **kwargs):
        wait = kwargs.get("wait", True)
        cmd_obj = self.get_command_object(command_name)
        return cmd_obj(*args, wait=wait)

    def init(self):
        self.ctrl = self.get_object_by_role("controller")
        self.ready_event = event.Event()
        if HWR.beamline.lims is None:
            self.log.warning(
                "EnergyScan: you should specify the database hardware object"
            )

    def is_connected(self):
        return True

    def get_static_parameters(self, config_file: str, element: str, edge: str) -> dict:
        """Get the static parameters for form the config file.
        Args:
            config_file(str): File to read the configuration from (full path).
            element(str): Element acronym as in the periodic table of elements.
            edge(str): edge line (K, L1, L2, L3)
        """
        pars = GetStaticParameters(config_file, element, edge).pars_dict

        offset_kev = self.get_property("offset_keV")
        pars["startEnergy"] += offset_kev
        pars["endEnergy"] += offset_kev
        pars["element"] = element

        return pars

    def escan_prepare(self):
        """Set the nesessary equipment in position for the scan."""
        self.ctrl.detcover.set_in()
        self.ctrl.diffractometer.fldet_in()
        self.ctrl.diffractometer.set_phase("DataCollection")

    def escan_postscan(self):
        """Actions done after the scan finished."""
        self.ctrl.diffractometer.fldet_out()

    def escan_cleanup(self):
        """Cleanup actions."""
        self.close_fast_shutter()
        HWR.beamline.safety_shutter.close()
        self.emit("energyScanFailed", ())
        self.ready_event.set()

    def close_fast_shutter(self):
        self.ctrl.diffractometer.msclose()

    def open_fast_shutter(self):
        self.ctrl.diffractometer.msopen()

    def cancelEnergyScan(self, *args):
        """Called by queue_entry.py. To be removed"""
        self.escan_cleanup()

    def store_energy_scan(self):
        if HWR.beamline.lims is None:
            return
        try:
            int(self.energy_scan_parameters["sessionId"])
        except (TypeError, KeyError):
            return

        # remove unnecessary for ISPyB fields:
        self.energy_scan_parameters.pop("prefix")
        self.energy_scan_parameters.pop("eroi_min")
        self.energy_scan_parameters.pop("eroi_max")
        self.energy_scan_parameters.pop("findattEnergy")
        self.energy_scan_parameters.pop("edge")
        self.energy_scan_parameters.pop("directory")
        self.energy_scan_parameters.pop("atomic_nb")

        spawn(store_energy_scan_thread, HWR.beamline.lims, self.energy_scan_parameters)

    def do_chooch(
        self, elt: str, edge: str, directory: str, archive_directory: str, prefix: str
    ):
        """Execute peak and IP calculation with chooch.
        Args:
           elt(str): Element acrony as in the periodic table of elements.
           edge(str): Edge like (K, L1, L2, L3).
           directory(str): raw data directory (fill path).
           archive_director(str): archive data directory (fill path).
           prefix(str): File root prefix.
        """
        self.energy_scan_parameters["endTime"] = time.strftime("%Y-%m-%d %H:%M:%S")

        raw_data_file = Path(directory) / "data.raw"

        archive_prefix = f"{prefix}_{elt}_{edge}"
        raw_scan_file = Path(directory) / (archive_prefix + ".raw")
        efs_scan_file = raw_scan_file.with_suffix(".efs")
        raw_arch_file = Path(archive_directory) / (archive_prefix + "1" + ".raw")

        i = 0
        while Path(raw_arch_file).is_file():
            i += 1
            raw_arch_file = Path(archive_directory) / (archive_prefix + str(i) + ".raw")

        png_scan_file = raw_scan_file.with_suffix(".png")
        png_arch_file = raw_arch_file.with_suffix(".png")

        if not Path(archive_directory).exists():
            Path(archive_directory).mkdir(parents=True)

        try:
            with Path(raw_scan_file).open("w") as fp:
                scan_data = []
                with Path(raw_data_file).open("r") as raw_file:
                    for line in raw_file.readlines()[2:]:
                        try:
                            x, y = line.split("\t")
                        except (AttributeError, ValueError):
                            x, y = line.split()
                        x = float(x.strip())
                        y = float(y.strip())
                        scan_data.append((x, y))
                        fp.write("%f,%f\r\n" % (x, y))
        except IOError:
            self.store_energy_scan()
            self.emit("energyScanFailed", ())
            return ()

        # create the gallery directory
        g_dir = Path(directory).parent / "gallery"
        if not Path(g_dir).exists():
            Path(g_dir).mkdir(parents=True)

        copy2(raw_scan_file, raw_arch_file)
        copy2(raw_scan_file, g_dir / raw_arch_file.name)
        self.energy_scan_parameters["scanFileFullPath"] = raw_arch_file

        # while waiting for chooch to work...
        subprocess.call(
            [
                "/cvmfs/sb.esrf.fr/bin/chooch",
                "-e",
                elt,
                "-a",
                edge,
                "-o",
                efs_scan_file,
                "-g",
                png_scan_file,
                raw_data_file,
            ]
        )
        time.sleep(5)
        with Path(efs_scan_file).open("r") as fp:
            for _ in range(3):
                next(fp)
            nparr = np.array([list(map(float, line.split())) for line in fp])
        fpp_peak = nparr[:, 1].max()
        idx = np.where(nparr[:, 1] == fpp_peak)
        pk = nparr[:, 0][idx][0] / 1000.0
        fp_peak = nparr[:, 2][idx][0]
        fpp_infl = nparr[:, 2].min()
        idx = np.where(nparr[:, 2] == fpp_infl)
        ip = nparr[:, 0][idx][0] / 1000.0
        fp_infl = nparr[:, 1][idx][0]
        # get the threshold from the theoretical edge [keV]
        th_t = self.get_property("theoritical_edge_threshold", 0.03)
        rm = pk + th_t

        th_edge = float(self.energy_scan_parameters["edgeEnergy"])

        msg = f"Chooch results: pk = {pk}, ip = {ip}. rm = {rm}.\n"
        msg += f"Theoretical edge: {th_edge}."
        self.log.info(msg)

        # +- shift from the theoretical edge [eV]
        edge_shift = 50
        calc_shift = (th_edge - pk) * 1000
        if abs(calc_shift) > edge_shift:
            rm = th_edge + th_t
            comm = "below" if calc_shift > edge_shift else "above"
            msg = f"Calculated peak {pk} is more than {edge_shift} eV {comm} "
            msg += f"the theoretical value {th_edge}. "
            self.energy_scan_parameters["comments"] = msg
            msg += "Check your scan and choose the energies manually"
            logging.getLogger("user_level_log").info(msg)
            pk = 0
            ip = 0

        efs_arch_file = raw_arch_file.with_suffix(".efs")
        if Path(efs_scan_file).is_file():
            copy2(efs_scan_file, efs_arch_file)
            copy2(efs_scan_file, g_dir / efs_arch_file.name)
        else:
            self.store_energy_scan()
            self.emit("energyScanFailed", ())
            return ()

        self.energy_scan_parameters["filename"] = raw_arch_file.name
        self.energy_scan_parameters["peakEnergy"] = pk
        self.energy_scan_parameters["inflectionEnergy"] = ip
        self.energy_scan_parameters["remoteEnergy"] = rm
        self.energy_scan_parameters["peakFPrime"] = fp_peak
        self.energy_scan_parameters["peakFDoublePrime"] = fpp_peak
        self.energy_scan_parameters["inflectionFPrime"] = fp_infl
        self.energy_scan_parameters["inflectionFDoublePrime"] = fpp_infl

        self.log.info("Saving png")
        # prepare to save png files
        title = "%10s  %6s  %6s\n%10s  %6.2f  %6.2f\n%10s  %6.2f  %6.2f" % (
            "energy",
            "f'",
            "f''",
            pk,
            fp_peak,
            fpp_peak,
            ip,
            fp_infl,
            fpp_infl,
        )
        if Path(png_scan_file).is_file():
            copy2(png_scan_file, png_arch_file)
            copy2(png_scan_file, g_dir / png_arch_file.name)
        else:
            self.store_energy_scan()
            self.emit("energyScanFailed", ())
            return ()

        self.energy_scan_parameters["jpegChoochFileFullPath"] = str(png_arch_file)
        self.store_energy_scan()

        self.emit(
            "chooch_finished",
            (
                pk,
                fpp_peak,
                fp_peak,
                ip,
                fpp_infl,
                fp_infl,
                rm,
                title,
            ),
        )
        return (
            pk,
            fpp_peak,
            fp_peak,
            ip,
            fpp_infl,
            fp_infl,
            rm,
            [],
            [],
            [],
            title,
        )

    def energy_scan_hook(self, energy_scan_parameters: dict):
        """
        Execute actions, required before running the raw scan(like changing
        undulator gaps, move to a given energy... These are in general
        beamline specific actions.
        """
        if self.energy_scan_parameters["findattEnergy"]:
            HWR.beamline.energy.set_value(energy_scan_parameters["findattEnergy"])

    def set_mca_roi(self, eroi_min, eroi_max):
        self.energy_scan_parameters["fluorescenceDetector"] = "KETEK_AXAS-A"
        # check if roi in eV or keV
        if eroi_min > 1000:
            eroi_min /= 1000.0
            eroi_max /= 1000.0
        self.ctrl.mca.set_roi(
            eroi_min,
            eroi_max,
            channel=1,
            element=self.energy_scan_parameters["element"],
            atomic_nb=self.energy_scan_parameters["atomic_nb"],
        )

    def choose_attenuation(self):
        """Choose the appropriate attenuation to execute the scan"""
        eroi_min = self.energy_scan_parameters["eroi_min"]
        eroi_max = self.energy_scan_parameters["eroi_max"]
        self.ctrl.detcover.set_in()
        mcafile = Path(self.energy_scan_parameters["directory"]) / "mca.raw"
        self.ctrl.find_max_attenuation(
            ctime=2, roi=[eroi_min, eroi_max], datafile=mcafile
        )
        self.energy_scan_parameters["transmissionFactor"] = (
            HWR.beamline.transmission.get_value()
        )

    def execute_energy_scan(self, energy_scan_parameters: dict):
        """
        Execute the raw scan sequence. Here is where you pass whatever
        parameters you need to run the raw scan (e.g start/end energy,
        counting time, energy step...).
        """
        start_en = energy_scan_parameters["startEnergy"]
        end_en = energy_scan_parameters["endEnergy"]
        dd = dt.now(tz=ZoneInfo("Europe/Paris"))
        fname = "%s/%s_%s_%s_%s.scan" % (
            energy_scan_parameters["directory"],
            energy_scan_parameters["prefix"],
            dt.datetime.strftime(dd, "%d"),
            dt.datetime.strftime(dd, "%b"),
            dt.datetime.strftime(dd, "%Y"),
        )
        self.ctrl.energy_scan.do_energy_scan(start_en, end_en, datafile=fname)
        self.energy_scan_parameters["exposureTime"] = (
            self.ctrl.energy_scan.exposure_time
        )


def store_energy_scan_thread(db_conn, scan_info):
    scan_info = dict(scan_info)
    blsample_id = scan_info["blSampleId"]
    scan_info.pop("blSampleId")

    try:
        db_status = db_conn.store_energy_scan(scan_info)
        if blsample_id is not None:
            try:
                escan_id = int(db_status["energyScanId"])
            except (NameError, KeyError):
                pass
            else:
                asso = {"blSampleId": blsample_id, "energyScanId": escan_id}
                db_conn.associate_bl_sample_and_energy_scan(asso)
    except Exception:
        logging.getLogger("HWR").exception("Could not store energy")



#  Project name: MXCuBE
#  https://github.com/mxcube
#
#  This file is part of MXCuBE software.
#
#  MXCuBE is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  MXCuBE is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with MXCuBE. If not, see <http://www.gnu.org/licenses/>.

import json
import logging
import os
import threading
import time
from copy import deepcopy
from queue import Queue

import cv2 as cv
import gevent
import Image
import numpy as np
import tine
from cStringIO import StringIO
from PIL.ImageQt import ImageQt
from scipy import (
    misc,
    ndimage,
)

from mxcubecore import HardwareRepository as HWR
from mxcubecore.HardwareObjects.abstract.AbstractCollect import AbstractCollect
from mxcubecore.HardwareObjects.QtGraphicsManager import QtGraphicsManager
from mxcubecore.model import queue_model_objects as qmo
from mxcubecore.TaskUtils import task
from mxcubecore.utils import (
    Colors,
    qt_import,
)

__credits__ = ["EMBL Hamburg"]
__category__ = "Task"


image_processing_queue = Queue()


def read_image(filename, timeout=10):
    if timeout:
        try:
            with gevent.Timeout(timeout):
                while not os.path.isfile(filename):
                    gevent.sleep(0.5)
                return cv.imread(filename, cv.IMREAD_ANYDEPTH)
        except gevent.Timeout:
            return
    else:
        return cv.imread(filename, cv.IMREAD_ANYDEPTH)


class EMBLXrayImaging(QtGraphicsManager, AbstractCollect):
    def __init__(self, *args):
        QtGraphicsManager.__init__(self, *args)
        AbstractCollect.__init__(self, *args)

        self.ff_apply = False
        self.ff_ssim = None
        self.qimage = None
        self.qpixmap = None
        self.image_count = 0
        self.image_reading_thread = None
        self.ff_corrected_list = []
        self.config_dict = {}
        self.collect_omega_start = 0
        self.omega_start = 0
        self.omega_move_enabled = False
        self.reference_distance = None
        self.reference_angle = None
        self.motor_positions = None

        self.image_dimension = (0, 0)
        self.graphics_camera_frame = None
        self.image_polling = None
        self.repeat_image_play = None
        self.current_image_index = None
        self.mouse_hold = False
        self.mouse_coord = [0, 0]
        self.centering_started = 0

        self._previous_collect_status = None
        self._actual_collect_status = None
        self._failed = False
        self._number_of_images = 0
        self.printed_warnings = []
        self.printed_errors = []

        self.chan_collect_status = None
        self.chan_collect_frame = None
        self.chan_collect_error = None
        self.chan_camera_error = None
        self.chan_camera_warning = None
        self.chan_ff_ssim = None

        self.cmd_collect_compression = None
        self.cmd_collect_detector = None
        self.cmd_collect_directory = None
        self.cmd_collect_exposure_time = None
        self.cmd_collect_in_queue = None
        self.cmd_collect_num_images = None
        self.cmd_collect_overlap = None
        self.cmd_collect_range = None
        self.cmd_collect_scan_type = None
        self.cmd_collect_shutter = None
        self.cmd_collect_start_angle = None
        self.cmd_collect_template = None
        self.cmd_collect_start = None
        self.cmd_collect_abort = None

        self.cmd_collect_ff_num_images = None
        self.cmd_collect_ff_offset = None
        self.cmd_collect_ff_pre = None
        self.cmd_collect_ff_post = None

        self.cmd_camera_trigger = None
        self.cmd_camera_live_view = None
        self.cmd_camera_write_data = None
        self.cmd_camera_ff_ssim = None

        self.beam_focusing_hwobj = None

    def init(self):
        AbstractCollect.init(self)
        self.ready_event = gevent.event.Event()
        self.image_dimension = (2048, 2048)
        self.reference_distance = self.get_property("reference_distance")
        self.reference_angle = self.get_property("reference_angle")

        QtGraphicsManager.init(self)

        self.disconnect(
            HWR.beamline.sample_view.camera, "imageReceived", self.camera_image_received
        )

        self.disconnect(
            HWR.beamline.diffractometer,
            "minidiffStateChanged",
            self.diffractometer_state_changed,
        )
        self.disconnect(
            HWR.beamline.diffractometer,
            "centringStarted",
            self.diffractometer_centring_started,
        )
        self.disconnect(
            HWR.beamline.diffractometer,
            "centringAccepted",
            self.create_centring_point,
        )
        self.disconnect(
            HWR.beamline.diffractometer,
            "centringSuccessful",
            self.diffractometer_centring_successful,
        )
        self.disconnect(
            HWR.beamline.diffractometer,
            "centringFailed",
            self.diffractometer_centring_failed,
        )
        self.disconnect(
            HWR.beamline.diffractometer,
            "pixelsPerMmChanged",
            self.diffractometer_pixels_per_mm_changed,
        )
        self.disconnect(
            HWR.beamline.diffractometer,
            "omegaReferenceChanged",
            self.diffractometer_omega_reference_changed,
        )
        self.disconnect(
            HWR.beamline.diffractometer,
            "minidiffPhaseChanged",
            self.diffractometer_phase_changed,
        )

        self.diffractometer_pixels_per_mm_changed((20.0, 20.0))
        self.graphics_manager_hwobj = self.get_object_by_role("graphics_manager")

        self.graphics_scale_item.set_start_position(20, self.image_dimension[1] - 20)

        self.graphics_scale_item.set_custom_pen_color(Colors.BLUE)
        self.graphics_omega_reference_item.set_custom_pen_color(Colors.DARK_BLUE)
        self.graphics_measure_distance_item.set_custom_pen_color(Colors.DARK_BLUE)
        self.graphics_beam_item.hide()

        self.graphics_view.scene().measureItemChanged.connect(self.measure_item_changed)
        self.graphics_view.scene().setSceneRect(
            0, 0, self.image_dimension[0], self.image_dimension[1]
        )

        self.qimage = qt_import.QImage()
        self.qpixmap = qt_import.QPixmap()

        self.chan_frame = self.get_channel_object("chanFrame")
        self.chan_frame.connect_signal("update", self.frame_changed)

        self.chan_ff_ssim = self.get_channel_object("chanFFSSIM")
        self.chan_ff_ssim.connect_signal("update", self.ff_ssim_changed)

        self.chan_collect_status = self.get_channel_object("collectStatus")
        # self._actual_collect_status = self.chan_collect_status.get_value()
        self.chan_collect_status.connect_signal("update", self.collect_status_update)

        self.chan_collect_frame = self.get_channel_object("chanFrameCount")
        self.chan_collect_frame.connect_signal("update", self.collect_frame_update)

        self.chan_collect_error = self.get_channel_object("collectError")
        self.chan_collect_error.connect_signal("update", self.collect_error_update)

        self.chan_camera_warning = self.get_channel_object("cameraWarning")
        self.chan_camera_warning.connect_signal("update", self.camera_warning_update)

        self.chan_camera_error = self.get_channel_object("cameraError")
        self.chan_camera_error.connect_signal("update", self.camera_error_update)

        self.cmd_collect_detector = self.get_command_object("collectDetector")
        self.cmd_collect_directory = self.get_command_object("collectDirectory")
        self.cmd_collect_exposure_time = self.get_command_object("collectExposureTime")
        self.cmd_collect_in_queue = self.get_command_object("collectInQueue")
        self.cmd_collect_num_images = self.get_command_object("collectNumImages")
        self.cmd_collect_range = self.get_command_object("collectRange")
        self.cmd_collect_scan_type = self.get_command_object("collectScanType")
        self.cmd_collect_shutter = self.get_command_object("collectShutter")
        self.cmd_collect_shutterless = self.get_command_object("collectShutterless")
        self.cmd_collect_start_angle = self.get_command_object("collectStartAngle")
        self.cmd_collect_template = self.get_command_object("collectTemplate")

        self.cmd_collect_ff_num_images = self.get_command_object("collectFFNumImages")
        self.cmd_collect_ff_offset = self.get_command_object("collectFFOffset")
        self.cmd_collect_ff_pre = self.get_command_object("collectFFPre")
        self.cmd_collect_ff_post = self.get_command_object("collectFFPost")

        self.cmd_camera_trigger = self.get_command_object("cameraTrigger")
        self.cmd_camera_live_view = self.get_command_object("cameraLiveView")
        self.cmd_camera_write_data = self.get_command_object("cameraWriteData")
        self.cmd_camera_ff_ssim = self.get_command_object("cameraFFSSIM")

        self.cmd_collect_start = self.get_command_object("collectStart")
        self.cmd_collect_abort = self.get_command_object("collectAbort")

        self.beam_focusing_hwobj = self.get_object_by_role("beam_focusing")

    def frame_changed(self, data):
        if self._collecting:
            jpgdata = StringIO(data)
            im = Image.open(jpgdata)
            self.qimage = ImageQt(im)
            self.graphics_camera_frame.setPixmap(
                self.qpixmap.fromImage(self.qimage, qt_import.Qt.MonoOnly)
            )

    def ff_ssim_changed(self, value):
        if self._collecting:
            self.ff_ssim = list(value)
            self.ff_ssim.sort()

    def mouse_clicked(self, pos_x, pos_y, left_click):
        QtGraphicsManager.mouse_clicked(self, pos_x, pos_y, left_click)
        if self.centering_started:
            self.motor_positions["phi"] = self.omega_angle
            HWR.beamline.diffractometer.set_static_positions(self.motor_positions)
            HWR.beamline.diffractometer.image_clicked(pos_x, pos_y)
            self.centering_started -= 1
        if not self.centering_started:
            self.set_centring_state(False)

    def mouse_released(self, pos_x, pos_y):
        QtGraphicsManager.mouse_released(self, pos_x, pos_y)
        self.mouse_hold = False

    def mouse_moved(self, pos_x, pos_y):
        QtGraphicsManager.mouse_moved(self, pos_x, pos_y)
        if self.mouse_hold:
            if self.mouse_coord[0] - pos_x > 0:
                index = self.current_image_index + 1
            elif self.mouse_coord[0] - pos_x < 0:
                index = self.current_image_index - 1
            else:
                return

            if index < 0:
                index = self.image_count - 1
            elif index >= self.image_count:
                index = 0
            self.mouse_coord[0] = pos_x
            self.display_image(index)

    def measure_item_changed(self, measured_points, measured_pix_num):
        start_x = measured_points[0].x()
        start_y = measured_points[0].y()
        end_x = measured_points[1].x()
        end_y = measured_points[1].y()

        if self.image_reading_thread is None:
            im = np.array(self.qimage.bits()).reshape(
                self.qimage.width(), self.qimage.height()
            )
        else:
            im = self.image_reading_thread.get_raw_image(self.current_image_index)
        # im_slice = im[start_x:start_y,end_x,end_y]
        # print im_slice.size, im_slice
        x = np.linspace(start_x, end_x, measured_pix_num)
        y = np.linspace(start_y, end_y, measured_pix_num)

        zi = ndimage.map_coordinates(im, np.vstack((x, y)))

        self.emit("measureItemChanged", zi)

    def get_graphics_view(self):
        return self.graphics_view

    def set_repeate_image_play(self, value):
        self.repeat_image_play = value

    def set_graphics_scene_size(self, size, fixed):
        pass

    def stop_move_beam_mark(self):
        """Stops to move beam mark

        :emits: infoMsg as str
        """
        self.set_cursor_busy(False)
        self.in_move_beam_mark_state = False
        self.graphics_move_beam_mark_item.hide()
        self.graphics_view.graphics_scene.update()
        pos_x = self.graphics_move_beam_mark_item.end_coord[0]
        pos_y = self.graphics_move_beam_mark_item.end_coord[1]

        HWR.beamline.diffractometer.set_imaging_beam_position(pos_x, pos_y)
        logging.getLogger("GUI").info(
            "Imaging beam position set to %d, %d" % (pos_x, pos_y)
        )
        self.emit("infoMsg", "")

    def diffractometer_phi_motor_moved(self, position):
        """Method when phi motor changed. Updates omega reference by
           redrawing phi angle

        :param position: phi rotation value
        :type position: float
        """
        QtGraphicsManager.diffractometer_phi_motor_moved(self, position)
        # logging.getLogger("GUI").info(str(position))
        self.display_image_by_angle(position)

    def pre_execute(self, data_model):
        self._failed = False
        """
        if self.beam_focusing_hwobj.get_focus_mode() != "imaging":
            self._error_msg = "Beamline is not in Imaging mode"
            self.emit("collectFailed", self._error_msg)
            logging.getLogger("GUI").error("Imaging: Error during acquisition (%s)" % self._error_msg)
            self.ready_event.set()
            self._collecting = False
            self._failed = True
            return
        """

        self.emit("progressInit", ("Image acquisition", 100, False))
        self._collect_frame = 0
        self.printed_warnings = []
        self.printed_errors = []
        self.ff_ssim = None
        self.config_dict = {}

        path_template = data_model.acquisitions[0].path_template
        acq_params = data_model.acquisitions[0].acquisition_parameters
        im_params = data_model.xray_imaging_parameters

        self._number_of_images = acq_params.num_images

        if im_params.detector_distance is not None:
            logging.getLogger("GUI").warning(
                "Imaging: Setting detector distance to %d mm"
                % int(im_params.detector_distance)
            )

            delta = (
                im_params.detector_distance - self.reference_distance
            ) * self.reference_angle
            for motor in self.beam_focusing_hwobj.get_focus_motors():
                if motor["motorName"] == "P14DetHor1":
                    target = motor["focusingModes"]["Imaging"] + delta
                    tine.set("/P14/P14DetTrans/P14detHor1", "Move.START", target)
                elif motor["motorName"] == "P14DetHor2":
                    target = motor["focusingModes"]["Imaging"] + delta
                    tine.set("/P14/P14DetTrans/P14detHor2", "Move.START", target)
            # TODO add later wait
            time.sleep(3)
            HWR.beamline.detector.distance.set_value(
                im_params.detector_distance, timeout=30
            )
            logging.getLogger("GUI").info("Imaging: Detector distance set")

        self.cmd_collect_detector("pco")
        self.cmd_collect_directory(str(path_template.directory))
        self.cmd_collect_template(str(path_template.get_image_file_name()))
        self.cmd_collect_scan_type("xrimg")

        self.cmd_collect_exposure_time(acq_params.exp_time)
        self.cmd_collect_num_images(acq_params.num_images)
        self.cmd_collect_start_angle(acq_params.osc_start)
        self.cmd_collect_range(acq_params.osc_range)
        self.cmd_collect_in_queue(acq_params.in_queue != False)
        shutter_name = HWR.beamline.detector.get_shutter_name()
        self.cmd_collect_shutter(shutter_name)

        self.cmd_collect_ff_num_images(im_params.ff_num_images)
        self.cmd_collect_ff_offset(
            [
                im_params.sample_offset_a,
                im_params.sample_offset_b,
                im_params.sample_offset_c,
            ]
        )
        self.cmd_collect_ff_pre(im_params.ff_pre)
        self.cmd_collect_ff_post(im_params.ff_post)

        if acq_params.osc_range == 0:
            self.cmd_camera_trigger(False)
        else:
            self.cmd_camera_trigger(True)
        self.cmd_camera_live_view(im_params.ff_apply)
        self.cmd_camera_write_data(im_params.camera_write_data)
        self.cmd_camera_ff_ssim(im_params.ff_ssim_enabled)

        self.set_osc_start(acq_params.osc_start)

        self.current_dc_parameters = qmo.to_collect_dict(
            data_model, HWR.beamline.session, qmo.Sample()
        )[0]
        self.current_dc_parameters["status"] = "Running"
        self.current_dc_parameters["comments"] = ""

        self.motor_positions = HWR.beamline.diffractometer.get_positions()
        self.take_crystal_snapshots()

        self.store_data_collection_in_lims()

    def execute(self, data_model):
        if not self._failed:
            self._collecting = True
            self.ready_event.clear()
            gevent.spawn(self.cmd_collect_start)
            # self.cmd_collect_start()
            # if data_model.xray_imaging_parameters.camera_write_data:
            #    self.read_images_task = gevent.spawn(self.load_images, None, None, None, data_model)
            self.ready_event.wait()
            self.ready_event.clear()

    def post_execute(self, data_model):
        self.emit("progressStop", ())
        self._collecting = False

        acq_params = data_model.acquisitions[0].acquisition_parameters
        path_template = data_model.acquisitions[0].path_template

        filename_template = "%s_%d_%0" + str(path_template.precision) + "d"
        config_filename = (
            filename_template
            % (
                path_template.base_prefix,
                path_template.run_number,
                path_template.start_num,
            )
            + ".json"
        )
        config_file_path = os.path.join(path_template.directory, config_filename)
        archive_config_path = os.path.join(
            path_template.get_archive_directory(), config_filename
        )

        self.config_dict = {
            "motor_pos": self.motor_positions,
            "collect": self.current_dc_parameters,
            "imaging": data_model.xray_imaging_parameters.as_dict(),
            "ff_ssim": None,
        }

        if data_model.xray_imaging_parameters.ff_pre:
            self.config_dict["ff_ssim"] = self.ff_ssim
        try:
            if not os.path.exists(path_template.get_archive_directory()):
                os.makedirs(path_template.get_archive_directory())
            with open(archive_config_path, "w") as outfile:
                json.dump(self.config_dict, outfile)
            logging.getLogger("GUI").info(
                "Imaging: Acquisition parameters saved in %s" % archive_config_path
            )
        except Exception:
            logging.getLogger("GUI").error(
                "Imaging: Unable to save acquisition parameters in %s"
                % archive_config_path
            )

        # self.current_dc_parameters["status"] = "Data collection successful"
        self._update_data_collection_in_lims()

        # Copy first and last image to ispyb
        if data_model.xray_imaging_parameters.camera_write_data:
            image_filename = (
                filename_template
                % (
                    path_template.base_prefix,
                    path_template.run_number,
                    path_template.start_num,
                )
                + ".jpeg"
            )
            image_filename = os.path.join(
                path_template.get_archive_directory(), image_filename
            )
            raw_image = read_image(
                path_template.get_image_path() % path_template.start_num, timeout=3
            )
            if raw_image is not None:
                misc.imsave(image_filename, raw_image)
                # Scale image from 2048x2048 to 256x256
                misc.imsave(
                    image_filename.replace(".jpeg", ".thumb.jpeg"),
                    misc.imresize(raw_image, (256, 256)),
                )
                self._store_image_in_lims(path_template.start_num)

                if acq_params.num_images > 1:
                    image_filename = (
                        filename_template
                        % (
                            path_template.base_prefix,
                            path_template.run_number,
                            acq_params.num_images - 1,
                        )
                        + ".jpeg"
                    )
                    image_filename = os.path.join(
                        path_template.get_archive_directory(), image_filename
                    )
                    raw_image = read_image(
                        path_template.get_image_path() % acq_params.num_images,
                        timeout=3,
                    )
                    if raw_image is not None:
                        misc.imsave(image_filename, raw_image)
                        misc.imsave(
                            image_filename.replace(".jpeg", ".thumb.jpeg"),
                            misc.imresize(raw_image, (256, 256)),
                        )
                        self._store_image_in_lims(
                            acq_params.num_images - path_template.start_num
                        )

    @task
    def _take_crystal_snapshot(self, filename):
        """Saves crystal snapshot"""
        self.graphics_manager_hwobj.save_scene_snapshot(filename)

    def data_collection_hook(self):
        pass

    @task
    def move_motors(self, motor_position_dict):
        """Move to centred position"""
        if motor_position_dict:
            HWR.beamline.diffractometer.move_motors(motor_position_dict)

    def trigger_auto_processing(self, process_event, frame_number):
        pass

    def collect_status_update(self, status):
        """Status event that controls execution

        :param status: collection status
        :type status: string
        """
        if status != self._actual_collect_status:
            self._previous_collect_status = self._actual_collect_status
            self._actual_collect_status = status
            if self._collecting:
                if self._actual_collect_status == "error":
                    # self.emit("collectFailed", self._error_msg)
                    error_msg = "Error during the acquisition (%s)" % self._error_msg
                    logging.getLogger("GUI").error("Imaging: %s" % error_msg)
                    self.collection_failed(error_msg)
                if self._previous_collect_status is None:
                    if self._actual_collect_status == "busy":
                        self.print_log(
                            "GUI", "info", "Imaging: Preparing acquisition..."
                        )
                elif self._previous_collect_status == "busy":
                    if self._actual_collect_status == "collecting":
                        self.emit("collectStarted", (None, 1))
                        self.print_log("GUI", "info", "Imaging: Acquisition started")
                    elif self._actual_collect_status == "ready":
                        self.ready_event.set()
                        self._collecting = False
                elif self._previous_collect_status == "collecting":
                    if self._actual_collect_status == "ready":
                        self.ready_event.set()
                        self._collecting = False
                        # if self.ff_ssim is None:
                        #    self.ff_ssim_changed(self.chan_ff_ssim.get_value())
                        logging.getLogger("GUI").info("Imaging: Acquisition done")
                    elif self._actual_collect_status == "aborting":
                        self.print_log("HWR", "info", "Imaging: Aborting...")
                        self.ready_event.set()
                        self._collecting = False

    def collect_error_update(self, error_msg):
        """Collect error behaviour

        :param error_msg: error message
        :type error_msg: string
        """

        if self._collecting and len(error_msg) > 0:
            self._error_msg = error_msg.replace("\n", "")
            logging.getLogger("GUI").error(
                "Imaging: Error from detector server: %s" % error_msg
            )

    def collect_frame_update(self, frame):
        """Image frame update

        :param frame: frame num
        :type frame: int
        """
        if self._collecting:
            if self._collect_frame != frame:
                self._collect_frame = frame
                self.emit(
                    "progressStep", (int(float(frame) / self._number_of_images * 100))
                )
                self.emit("collectImageTaken", frame)

    def camera_warning_update(self, warning_str):
        if self._collecting:
            if warning_str.endswith("\n"):
                warning_str = warning_str[:-1]
            if warning_str.startswith("\n"):
                warning_str = warning_str[1:]
            warning_list = warning_str.split("\n")

            for warning in warning_list:
                if warning and warning not in self.printed_warnings:
                    logging.getLogger("GUI").warning(
                        "Imaging: PCO camera warning: %s" % warning
                    )
                    self.printed_warnings.append(warning)

    def camera_error_update(self, error_str):
        if self._collecting:
            if error_str.endswith("\n"):
                error_str = error_str[:-1]
            if error_str.startswith("\n"):
                error_str = error_str[1:]
            error_list = error_str.split("\n")

            for error in error_list:
                if error and error not in self.printed_errors:
                    logging.getLogger("GUI").error(
                        "Imaging: PCO camera error: %s" % error
                    )
                    self.printed_errors.append(error)

    def set_ff_apply(self, state):
        self.ff_apply = state
        self.display_image(self.current_image_index)

    def display_image_by_angle(self, angle=None):
        if self.config_dict:
            if not angle:
                angle = self.omega_angle
            osc_seq = self.config_dict["collect"]["oscillation_sequence"][0]
            index = int(osc_seq["range"] * (angle - osc_seq["start"]))
            self.display_image(index)

    def display_image(self, index):
        if self.image_reading_thread is None:
            return

        # osc_seq = self.config_dict["collect"]["oscillation_sequence"][0]
        # angle = osc_seq["start"] + index * osc_seq["range"]
        # self.motor_positions["phi"] = angle
        # HWR.beamline.diffractometer.set_static_positions(self.motor_positions)
        # self.graphics_omega_reference_item.set_phi_position(angle)
        self.current_image_index = index

        raw_image = self.image_reading_thread.get_raw_image(index)

        if self.ff_apply:
            if self.ff_corrected_list[index] is None:
                corrected_im_min_max = (
                    self.image_reading_thread.get_corrected_im_min_max()
                )
                ff_image = self.image_reading_thread.get_ff_image(index).astype(float)
                ff_corrected_image = np.divide(
                    raw_image.astype(float),
                    ff_image,
                    out=np.ones_like(raw_image.astype(float)),
                    where=ff_image != 0,
                )
                im = (
                    255.0
                    * (ff_corrected_image - corrected_im_min_max[0])
                    / (corrected_im_min_max[1] - corrected_im_min_max[0])
                )
                self.ff_corrected_list[index] = im.astype(np.uint16)
            else:
                im = self.ff_corrected_list[index]
        else:
            raw_im_min_max = self.image_reading_thread.get_raw_im_min_max()
            im = (
                255.0
                * (raw_image - raw_im_min_max[0])
                / (raw_im_min_max[1] - raw_im_min_max[0])
            )

        if im is not None:
            self.qimage = qt_import.QImage(
                im.astype(np.uint8),
                im.shape[1],
                im.shape[0],
                im.shape[1],
                qt_import.QImage.Format_Indexed8,
            )
            self.graphics_camera_frame.setPixmap(self.qpixmap.fromImage(self.qimage))
            self.emit("imageLoaded", index)

    def display_image_relative(self, relative_index):
        self.display_image(self.current_image_index + relative_index)

    def play_image_relative(self, relative_angle):
        self.play_images(0.04, relative_angle, False)

    def set_osc_start(self, osc_start):
        self.collect_omega_start = osc_start

    def set_omega_move_enabled(self, state):
        self.omega_move_enabled = state

    def get_ff_and_config_path(self, raw_image_path):
        directory = os.path.dirname(raw_image_path)
        filename = os.path.basename(raw_image_path)

        ff_path = os.path.join(directory, "ff_%s_00001.tiff" % filename[:-11])
        if not os.path.exists(ff_path):
            ff_path = None
        config_path = os.path.join(
            directory.replace("mnt/beegfs/P14", "data/ispyb/p14"),
            "%s_00001.json" % filename[:-11],
        )
        if not os.path.exists(config_path):
            config_path = None

        return ff_path, config_path

    def load_images(
        self,
        data_path=None,
        flat_field_path=None,
        config_path=None,
        data_model=None,
        load_all=True,
    ):
        ff_ssim = None
        raw_filename_list = []
        ff_filename_list = []
        self.config_dict = {}
        self.omega_start = HWR.beamline.diffractometer.get_omega_position()
        self.motor_positions = None
        self.image_reading_thread = None

        if not data_model:
            if data_path.endswith("tiff"):
                ext_len = 4
            else:
                ext_len = 3

            base_name_list = os.path.splitext(os.path.basename(data_path))
            prefix = base_name_list[0][: -(ext_len + 1)]

            # Reading config json --------------------------------------------
            if config_path is None:
                config_path = data_path[:-ext_len] + "json"
            if os.path.exists(config_path):
                with open(config_path) as f:
                    self.config_dict = json.load(f)
                ff_ssim = self.config_dict["ff_ssim"]
                self.motor_positions = deepcopy(self.config_dict["motor_pos"])
                HWR.beamline.diffractometer.set_static_positions(self.motor_positions)
            else:
                logging.getLogger("user_level_log").error(
                    "Imaging: Unable to open config file %s" % config_path
                )

        if data_model:
            if data_model.xray_imaging_parameters.ff_pre:
                acq_params = data_model.acquisitions[0].acquisition_parameters
                path_template = data_model.acquisitions[0].path_template
                for index in range(data_model.xray_imaging_parameters.ff_num_images):
                    ff_filename_list.append(
                        os.path.join(
                            path_template.directory,
                            "ff_" + path_template.get_image_file_name() % (index + 1),
                        )
                    )
        elif os.path.exists(flat_field_path):
            base_name_list = os.path.splitext(os.path.basename(data_path))
            ff_prefix = base_name_list[0][: -(ext_len + 1)]
            os.chdir(os.path.dirname(flat_field_path))
            ff_filename_list = sorted(
                [
                    os.path.join(os.path.dirname(flat_field_path), f)
                    for f in os.listdir(os.path.dirname(flat_field_path))
                    if f.startswith("ff_" + prefix)
                ]
            )

        # Reading raw images -------------------------------------------------
        if data_model:
            acq_params = data_model.acquisitions[0].acquisition_parameters
            path_template = data_model.acquisitions[0].path_template
            for index in range(acq_params.num_images):
                raw_filename_list.append(
                    os.path.join(
                        path_template.directory,
                        path_template.get_image_file_name() % (index + 1),
                    )
                )
        elif os.path.exists(data_path):
            os.chdir(os.path.dirname(data_path))
            raw_filename_list = sorted(
                [
                    os.path.join(os.path.dirname(data_path), f)
                    for f in os.listdir(os.path.dirname(data_path))
                    if f.startswith(prefix)
                ]
            )
        else:
            acq_params = data_model.acquisitions[0].acquisition_parameters
            path_template = data_model.acquisitions[0].path_template
            for index in range(acq_params.num_images):
                raw_filename_list.append(
                    os.path.join(
                        path_template.directory,
                        path_template.get_image_file_name() % (index + 1),
                    )
                )

        self.image_count = len(raw_filename_list)
        if self.image_reading_thread is not None:
            self.image_reading_thread.set_stop()

        self.ff_corrected_list = [None] * self.image_count

        self.image_reading_thread = ImageReadingThread(
            raw_filename_list, ff_filename_list, ff_ssim
        )
        self.image_reading_thread.start()

        self.current_image_index = 0
        self.emit("imageInit", self.image_count)

        gevent.sleep(2)
        self.last_image_index = 0
        self.display_image_by_angle()

    def play_images(self, exp_time=0.04, relative_angle=None, repeat=True):
        self.image_polling = gevent.spawn(
            self.do_image_polling, exp_time, relative_angle, repeat
        )

    def do_image_polling(self, exp_time=0.04, relative_angle=0, repeat=False):
        self.repeat_image_play = repeat

        direction = 1
        if relative_angle < 0:
            direction = -1
        rotate_in_sec = 2
        step = int(self.image_count / (rotate_in_sec / exp_time))
        start_index = self.current_image_index
        index = 0

        while self.current_image_index < self.image_count:
            if relative_angle:
                if index >= abs(self.image_count / 360.0 * relative_angle):
                    break
            self.log.debug("display: " + str(self.current_image_index))
            self.display_image(self.current_image_index)
            self.current_image_index += direction * step
            if self.repeat_image_play and self.current_image_index >= self.image_count:
                self.current_image_index = 0
            gevent.sleep(exp_time)
            index += step
        if relative_angle:
            self.display_image(
                int(start_index + self.image_count / 360.0 * relative_angle)
            )

    def stop_image_play(self):
        self.image_polling.kill()

    def stop_collect(self):
        self.cmd_collect_abort()

    def mouse_wheel_scrolled(self, delta):
        if (
            self.image_reading_thread is None
            or self.image_reading_thread.get_raw_image(self.current_image_index) is None
        ):
            return

        if delta > 0:
            self.current_image_index -= 1
            if self.current_image_index < 0:
                self.current_image_index = self.image_count - 1
        else:
            self.current_image_index += 1
            if self.current_image_index == self.image_count:
                self.current_image_index = 0
        self.display_image(self.current_image_index)

    def start_centering(self):
        self.centering_started = 3
        self.set_centring_state(True)

        # osc_seq = self.config_dict["collect"]["oscillation_sequence"][0]
        # angle = osc_seq["start"] + index * osc_seq["range"]
        self.motor_positions["phi"] = self.omega_angle
        HWR.beamline.diffractometer.set_static_positions(self.motor_positions)

        HWR.beamline.diffractometer.start_centring_method(
            HWR.beamline.diffractometer.CENTRING_METHOD_IMAGING
        )

    def start_n_centering(self):
        self.centering_started = 100
        self.set_centring_state(True)
        HWR.beamline.diffractometer.start_centring_method(
            HWR.beamline.diffractometer.CENTRING_METHOD_IMAGING_N
        )


class GraphicsCameraFrame(qt_import.QGraphicsPixmapItem):
    def __init__(self, parent=None):
        super(GraphicsCameraFrame, self).__init__(parent)

    def mousePressEvent(self, event):
        pos = qt_import.QPointF(event.pos())
        self.scene().parent().mouseClickedSignal.emit(
            pos.x(), pos.y(), event.button() == qt_import.Qt.LeftButton
        )
        self.update()

    def mouseMoveEvent(self, event):
        pos = qt_import.QPointF(event.pos())
        self.scene().parent().mouseMovedSignal.emit(pos.x(), pos.y())
        self.update()

    # def mouseDoubleClickEvent(self, event):
    #    position = qt_import.QPointF(event.pos())
    #    self.scene().mouseDoubleClickedSignal.emit(position.x(), position.y())
    #    self.update()

    def mouseReleaseEvent(self, event):
        pos = qt_import.QPointF(event.pos())
        self.scene().parent().mouseReleasedSignal.emit(pos.x(), pos.y())
        self.update()


class GraphicsView(qt_import.QGraphicsView):
    mouseClickedSignal = qt_import.pyqtSignal(int, int, bool)
    mouseReleasedSignal = qt_import.pyqtSignal(int, int)
    mouseMovedSignal = qt_import.pyqtSignal(int, int)
    keyPressedSignal = qt_import.pyqtSignal(str)
    wheelSignal = qt_import.pyqtSignal(int)

    def __init__(self, parent=None):
        super(GraphicsView, self).__init__(parent)

        self.setScene(qt_import.QGraphicsScene(self))
        self.scene().clearSelection()
        self.setMouseTracking(True)
        self.setDragMode(qt_import.QGraphicsView.RubberBandDrag)
        self.setHorizontalScrollBarPolicy(qt_import.Qt.ScrollBarAlwaysOff)
        self.setVerticalScrollBarPolicy(qt_import.Qt.ScrollBarAlwaysOff)

    def wheelEvent(self, event):
        self.wheelSignal.emit(event.delta())


class ImageReadingThread(threading.Thread):
    def __init__(self, raw_filename_list, ff_filename_list=[], ff_ssim=[]):
        threading.Thread.__init__(self)

        self.stopped = False
        self.failed_to_read_count = 10
        self.raw_filename_list = raw_filename_list
        self.ff_filename_list = ff_filename_list

        self.raw_image_list = [None] * len(self.raw_filename_list)
        self.ff_image_list = [None] * len(self.ff_filename_list)

        self.ff_ssim = None
        self.raw_im_min_max = [pow(2, 16), 0]
        self.corrected_im_min_max = [pow(2, 16), 0]

    def start(self):
        # self.thread_watcher = gevent.get_hub().loop.async()
        self.thread_done = gevent.event.Event()
        threading.Thread.start(self)
        return self.thread_done

    def set_stop(self):
        self.stopped = True
        logging.getLogger("GUI").info("Image reading stopped")

    def set_ff_ssim(self, ff_ssim):
        self.ff_ssim = ff_ssim

    def run(self):
        logging.getLogger("GUI").info("Image reading started...")
        for index, filename in enumerate(self.ff_filename_list):
            if self.stopped:
                # self.thread_watcher.send()
                return
            self.ff_image_list[index] = read_image(filename)

        progress_step = 20

        for index, filename in enumerate(self.raw_filename_list):
            if self.stopped:
                return

            raw_image = read_image(filename)
            self.raw_image_list[index] = raw_image

            if index == 0:
                self.raw_im_min_max[0] = raw_image[8:].min()
                self.raw_im_min_max[1] = raw_image[8:].max()

            if self.ff_filename_list and index == 0:
                ff_image = self.get_ff_image(index)
                ff_applied = np.divide(
                    raw_image.astype(float),
                    ff_image.astype(float),
                    out=np.ones_like(raw_image.astype(float)),
                    where=ff_image.astype(float) != 0,
                )
                ff_applied[ff_image == (pow(2, 16) - 1)] = 1

                self.corrected_im_min_max[0] = ff_applied[8:].min()
                self.corrected_im_min_max[1] = ff_applied[8:].max()

            done_per = int(float(index) / len(self.raw_filename_list) * 100)
            if (
                not index % (len(self.raw_filename_list) / (100 / progress_step))
                and done_per > 0
            ):
                logging.getLogger("GUI").info("Image reading %d%% completed" % done_per)

        logging.getLogger("GUI").info("Image reading finished")
        # self.thread_watcher.send()

    def get_raw_image(self, index):
        return self.raw_image_list[index]

    def get_raw_im_min_max(self):
        return self.raw_im_min_max

    def get_corrected_im_min_max(self):
        return self.corrected_im_min_max

    def get_ff_image(self, raw_image_index):
        print("get_ff_image ", raw_image_index)
        if self.ff_ssim:
            ff_index = self.ff_ssim[raw_image_index][2] - 1
        else:
            ff_index = int(
                raw_image_index
                / float(len(self.raw_image_list))
                * len(self.ff_image_list)
            )
        print(ff_index)
        return self.ff_image_list[ff_index]



import json
import pathlib

from mxcubecore import HardwareRepository as HWR
from mxcubecore.HardwareObjects.ICATLIMS import ICATLIMS


class SSXICATLIMS(ICATLIMS):
    """
    ICAT+ client for SSX.
    """

    def store_data_collection(self, parameters, bl_config=None):
        pass

    def update_data_collection(self, parameters):
        pass

    def finalize_data_collection(self, parameters):
        self.log.info("Storing data to ICAT")
        collection_parameters = parameters["collection_parameters"]
        beamline_parameters = parameters["beamline_parameters"]
        data_path = parameters["data_path"]
        extra_lims_values = parameters["extra_lims_values"]

        try:
            data = {
                "MX_scanType": "SSX-Jet",
                "MX_beamShape": beamline_parameters.beam_shape,
                "MX_beamSizeAtSampleX": beamline_parameters.beam_size_x,
                "MX_beamSizeAtSampleY": beamline_parameters.beam_size_y,
                "MX_detectorDistance": beamline_parameters.detector_distance,
                "MX_directory": data_path,
                "MX_exposureTime": (
                    collection_parameters.user_collection_parameters.exp_time
                ),
                "MX_flux": extra_lims_values.flux_start,
                "MX_fluxEnd": extra_lims_values.flux_end,
                "MX_numberOfImages": (
                    collection_parameters.collection_parameters.num_images
                ),
                "MX_resolution": beamline_parameters.resolution,
                "MX_transmission": beamline_parameters.transmission,
                "MX_xBeam": beamline_parameters.beam_x,
                "MX_yBeam": beamline_parameters.beam_y,
                "Sample_name": collection_parameters.path_parameters.prefix,
                "InstrumentMonochromator_wavelength": beamline_parameters.wavelength,
                "chipModel": extra_lims_values.chip_model,
                "monoStripe": extra_lims_values.mono_stripe,
                "energyBandwidth": beamline_parameters.energy_bandwidth,
                "detector_id": HWR.beamline.detector.get_property("detector_id"),
                "experimentType": collection_parameters.common_parameters.type,
            }

            data.update(collection_parameters.user_collection_parameters.dict())
            data.update(collection_parameters.collection_parameters.dict())

            self.icatClient.store_dataset(
                beamline="ID29",
                proposal=f"{HWR.beamline.session.proposal_code}{HWR.beamline.session.proposal_number}",
                dataset=collection_parameters.path_parameters.prefix,
                path=data_path,
                metadata=data,
            )

            icat_metadata_path = pathlib.Path(data_path) / "metadata.json"
            with open(icat_metadata_path, "w") as f:
                f.write(json.dumps(data, indent=4))
                self.log.info(f"Wrote {icat_metadata_path}")

        except Exception as e:
            self.log.exception("Failed uploading to ICAT (%s)", e)
